from unittest import mock

import pytest
from utils import SocketMock

from ophyd_devices.smaract import SmaractController
from ophyd_devices.smaract.smaract_controller import SmaractCommunicationMode
from ophyd_devices.smaract.smaract_errors import SmaractCommunicationError, SmaractErrorCode
from ophyd_devices.smaract.smaract_ophyd import SmaractMotor


@pytest.fixture
def controller():
    SmaractController._reset_controller()
    controller = SmaractController(socket_cls=SocketMock, socket_host="dummy", socket_port=123)
    controller.on()
    controller.sock.flush_buffer()
    yield controller


@pytest.fixture
def lsmarA():
    SmaractController._reset_controller()
    motor_a = SmaractMotor(
        "A", name="lsmarA", host="mpc2680.psi.ch", port=8085, sign=1, socket_cls=SocketMock
    )
    motor_a.controller.on()
    motor_a.controller.sock.flush_buffer()
    motor_a.stage()
    yield motor_a


@pytest.mark.parametrize(
    "axis,position,get_message,return_msg",
    [
        (0, 50, b":GP0\n", b":P0,50000000\n"),
        (1, 0, b":GP1\n", b":P1,0\n"),
        (0, -50, b":GP0\n", b":P0,-50000000\n"),
        (0, -50.23, b":GP0\n", b":P0,-50230000\n"),
    ],
)
def test_get_position(controller, axis, position, get_message, return_msg):
    controller.sock.buffer_recv = return_msg
    val = controller.get_position(axis)
    assert val == position
    assert controller.sock.buffer_put[0] == get_message


@pytest.mark.parametrize(
    "axis,is_referenced,get_message,return_msg,exception",
    [
        (0, True, b":GPPK0\n", b":PPK0,1\n", None),
        (1, True, b":GPPK1\n", b":PPK1,1\n", None),
        (0, False, b":GPPK0\n", b":PPK0,0\n", None),
        (200, False, b":GPPK0\n", b":PPK0,0\n", ValueError),
    ],
)
def test_axis_is_referenced(controller, axis, is_referenced, get_message, return_msg, exception):
    controller.sock.buffer_recv = return_msg
    if exception is not None:
        with pytest.raises(exception):
            val = controller.axis_is_referenced(axis)
    else:
        val = controller.axis_is_referenced(axis)
        assert val == is_referenced
        assert controller.sock.buffer_put[0] == get_message


@pytest.mark.parametrize(
    "return_msg,exception,raised",
    [
        (b"false\n", SmaractCommunicationError, False),
        (b":E0,1", SmaractErrorCode, True),
        (b":E,1", SmaractCommunicationError, True),
        (b":E,-1", SmaractCommunicationError, True),
    ],
)
def test_socket_put_and_receive_raises_exception(controller, return_msg, exception, raised):
    controller.sock.buffer_recv = return_msg
    with pytest.raises(exception):
        controller.socket_put_and_receive(b"test", raise_if_not_status=True)

    controller.sock.flush_buffer()
    controller.sock.buffer_recv = return_msg

    if raised:
        with pytest.raises(exception):
            controller.socket_put_and_receive(b"test")
    else:
        assert controller.socket_put_and_receive(b"test") == return_msg.split(b"\n")[0].decode()


@pytest.mark.parametrize(
    "mode,get_message,return_msg", [(0, b":GCM\n", b":CM0\n"), (1, b":GCM\n", b":CM1\n")]
)
def test_communication_mode(controller, mode, get_message, return_msg):
    controller.sock.buffer_recv = return_msg
    val = controller.get_communication_mode()
    assert controller.sock.buffer_put[0] == get_message
    assert val == SmaractCommunicationMode(mode)


@pytest.mark.parametrize(
    "is_moving,get_message,return_msg",
    [
        (0, b":GS0\n", b":S0,0\n"),
        (1, b":GS0\n", b":S0,1\n"),
        (1, b":GS0\n", b":S0,2\n"),
        (0, b":GS0\n", b":S0,3\n"),
        (1, b":GS0\n", b":S0,4\n"),
        (0, b":GS0\n", b":S0,5\n"),
        (0, b":GS0\n", b":S0,6\n"),
        (1, b":GS0\n", b":S0,7\n"),
        (0, b":GS0\n", b":S0,9\n"),
        (0, [b":GS0\n", b":GS0\n"], [b":E0,0\n", b":S0,9"]),
    ],
)
def test_axis_is_moving(controller, is_moving, get_message, return_msg):
    controller.sock.buffer_recv = return_msg
    val = controller.is_axis_moving(0)
    assert val == is_moving
    if isinstance(controller.sock.buffer_put, list) and len(controller.sock.buffer_put) == 1:
        controller.sock.buffer_put = controller.sock.buffer_put[0]
    assert controller.sock.buffer_put == get_message


@pytest.mark.parametrize(
    "sensor_id,axis,get_msg,return_msg",
    [
        (1, 0, b":GST0\n", b":ST0,1\n"),
        (6, 0, b":GST0\n", b":ST0,6\n"),
        (6, 1, b":GST1\n", b":ST1,6\n"),
    ],
)
def test_get_sensor_definition(controller, sensor_id, axis, get_msg, return_msg):
    controller.sock.buffer_recv = return_msg
    sensor = controller.get_sensor_type(axis)
    assert sensor.type_code == sensor_id


@pytest.mark.parametrize(
    "move_speed,axis,get_msg,return_msg",
    [
        (50, 0, b":SCLS0,50000000\n", b":E-1,0"),
        (0, 0, b":SCLS0,0\n", b":E-1,0"),
        (20.23, 1, b":SCLS1,20230000\n", b":E-1,0"),
    ],
)
def test_set_move_speed(controller, move_speed, axis, get_msg, return_msg):
    controller.sock.buffer_recv = return_msg
    controller.set_closed_loop_move_speed(axis, move_speed)
    assert controller.sock.buffer_put[0] == get_msg


@pytest.mark.parametrize(
    "pos,axis,hold_time,get_msg,return_msg",
    [
        (50, 0, None, b":MPA0,50000000,1000\n", b":E0,0"),
        (0, 0, 800, b":MPA0,0,800\n", b":E0,0"),
        (20.23, 1, None, b":MPA1,20230000,1000\n", b":E0,0"),
    ],
)
def test_move_axis_to_absolute_position(controller, pos, axis, hold_time, get_msg, return_msg):
    controller.sock.buffer_recv = return_msg
    if hold_time is not None:
        controller.move_axis_to_absolute_position(axis, pos, hold_time=hold_time)
    else:
        controller.move_axis_to_absolute_position(axis, pos)
    assert controller.sock.buffer_put[0] == get_msg


@pytest.mark.parametrize(
    "pos,get_msg,return_msg",
    [
        (
            50,
            [b":GPPK0\n", b":MPA0,50000000,1000\n", b":GS0\n", b":GP0\n"],
            [b":PPK0,1\n", b":E0,0\n", b":S0,0\n", b":P0,50000000\n"],
        ),
        (
            0,
            [b":GPPK0\n", b":MPA0,0,1000\n", b":GS0\n", b":GP0\n"],
            [b":PPK0,1\n", b":E0,0\n", b":S0,0\n", b":P0,0000000\n"],
        ),
        (
            20.23,
            [b":GPPK0\n", b":MPA0,20230000,1000\n", b":GS0\n", b":GP0\n"],
            [b":PPK0,1\n", b":E0,0\n", b":S0,0\n", b":P0,20230000\n"],
        ),
        (
            20.23,
            [b":GPPK0\n", b":GPPK0\n", b":MPA0,20230000,1000\n", b":GS0\n", b":GP0\n"],
            [b":S0,0\n", b":PPK0,1\n", b":E0,0\n", b":S0,0\n", b":P0,20230000\n"],
        ),
    ],
)
def test_move_axis(lsmarA, pos, get_msg, return_msg):
    controller = lsmarA.controller
    controller.sock.buffer_recv = return_msg
    lsmarA.move(pos)
    assert controller.sock.buffer_put == get_msg


@pytest.mark.parametrize("num_axes,get_msg,return_msg", [(1, [b":S0\n"], [b":E0,0"])])
def test_stop_axis(lsmarA, num_axes, get_msg, return_msg):
    controller = lsmarA.controller
    controller.sock.buffer_recv = return_msg
    controller.stop_all_axes()
    assert controller.sock.buffer_put == get_msg


def test_all_axes_referenced(lsmarA):
    controller = lsmarA.controller
    with mock.patch.object(controller, "axis_is_referenced", return_value=True) as mock_is_ref:
        val = controller.all_axes_referenced()
        assert val
        mock_is_ref.assert_called_once_with(0)



from unittest import mock

from bec_lib.devicemanager import DeviceContainer
from bec_lib.tests.utils import ConnectorMock


class SocketMock:
    """Socket Mock. Used for testing"""

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.buffer_put = []
        self.buffer_recv = [b""]
        self.is_open = False
        self.sock = None
        self.open()

    def connect(self):
        """Mock connect method"""
        print(f"connecting to {self.host} port {self.port}")

    def _put(self, msg_bytes):
        """Mock put method"""
        self.buffer_put.append(msg_bytes)
        print(self.buffer_put)

    # pylint: disable=unused-argument
    def _recv(self, buffer_length=1024):
        """Mock receive method"""
        print(self.buffer_recv)
        if isinstance(self.buffer_recv, list):
            if len(self.buffer_recv) > 0:
                ret_val = self.buffer_recv.pop(0)
            else:
                ret_val = b""
            return ret_val
        return self.buffer_recv

    def _initialize_socket(self):
        """Mock initialize socket method"""

    def put(self, msg):
        """Mock put method"""
        return self._put(msg)

    def receive(self, buffer_length=1024):
        """Mock receive method"""
        return self._recv(buffer_length=buffer_length)

    def open(self):
        """Mock open method"""
        self._initialize_socket()
        self.is_open = True

    def close(self):
        """Mock close method"""
        self.sock = None
        self.is_open = False

    def flush_buffer(self):
        """Mock flush buffer method"""
        self.buffer_put = []
        self.buffer_recv = ""


class MockPV:
    """
    MockPV class

    This class is used for mocking pyepics signals for testing purposes

    """

    _fmtsca = "<PV '%(pvname)s', count=%(count)i, type=%(typefull)s, access=%(access)s>"
    _fmtarr = "<PV '%(pvname)s', count=%(count)i/%(nelm)i, type=%(typefull)s, access=%(access)s>"
    _fields = (
        "pvname",
        "value",
        "char_value",
        "status",
        "ftype",
        "chid",
        "host",
        "count",
        "access",
        "write_access",
        "read_access",
        "severity",
        "timestamp",
        "posixseconds",
        "nanoseconds",
        "precision",
        "units",
        "enum_strs",
        "upper_disp_limit",
        "lower_disp_limit",
        "upper_alarm_limit",
        "lower_alarm_limit",
        "lower_warning_limit",
        "upper_warning_limit",
        "upper_ctrl_limit",
        "lower_ctrl_limit",
    )

    def __init__(
        self,
        pvname,
        callback=None,
        form="time",
        verbose=False,
        auto_monitor=None,
        count=None,
        connection_callback=None,
        connection_timeout=None,
        access_callback=None,
    ):
        self.pvname = pvname.strip()
        self.form = form.lower()
        self.verbose = verbose
        self._auto_monitor = auto_monitor
        self.ftype = None
        self.connected = True
        self.connection_timeout = connection_timeout
        self._user_max_count = count

        if self.connection_timeout is None:
            self.connection_timeout = 3
        self._args = {}.fromkeys(self._fields)
        self._args["pvname"] = self.pvname
        self._args["count"] = count
        self._args["nelm"] = -1
        self._args["type"] = "unknown"
        self._args["typefull"] = "unknown"
        self._args["access"] = "unknown"
        self._args["status"] = 0
        self.connection_callbacks = []
        self.mock_data = 0

        if connection_callback is not None:
            self.connection_callbacks = [connection_callback]

        self.access_callbacks = []
        if access_callback is not None:
            self.access_callbacks = [access_callback]

        self.callbacks = {}
        self._put_complete = None
        self._monref = None  # holder of data returned from create_subscription
        self._monref_mask = None
        self._conn_started = False
        if isinstance(callback, (tuple, list)):
            for i, thiscb in enumerate(callback):
                if callable(thiscb):
                    self.callbacks[i] = (thiscb, {})
        elif callable(callback):
            self.callbacks[0] = (callback, {})

        self.chid = None
        self.context = mock.MagicMock()
        self._cache_key = (pvname, form, self.context)
        self._reference_count = 0
        for conn_cb in self.connection_callbacks:
            conn_cb(pvname=pvname, conn=True, pv=self)
        for acc_cb in self.access_callbacks:
            acc_cb(True, True, pv=self)

    # pylint disable: unused-argument
    def wait_for_connection(self, timeout=None):
        """Wait for connection"""
        return self.connected

    # pylint disable: unused-argument
    def get_all_metadata_blocking(self, timeout):
        """Get all metadata blocking"""
        md = self._args.copy()
        md.pop("value", None)
        return md

    def get_all_metadata_callback(self, callback, *, timeout):
        """Get all metadata callback"""

        def get_metadata_thread(pvname):
            md = self.get_all_metadata_blocking(timeout=timeout)
            callback(pvname, md)

        get_metadata_thread(pvname=self.pvname)

    # pylint disable: unused-argument
    def put(
        self, value, wait=False, timeout=None, use_complete=False, callback=None, callback_data=None
    ):
        """MOCK PV, put function"""
        self.mock_data = value
        if callback is not None:
            callback(None, None, None)

    # pylint: disable=unused-argument
    def add_callback(self, callback=None, index=None, run_now=False, with_ctrlvars=True, **kw):
        """Add callback"""
        return mock.MagicMock()

    # pylint: disable=unused-argument
    def get_with_metadata(
        self,
        count=None,
        as_string=False,
        as_numpy=True,
        timeout=None,
        with_ctrlvars=False,
        form=None,
        use_monitor=True,
        as_namespace=False,
    ):
        """Get MOCKPV data together with metadata"""
        return {"value": self.mock_data}

    def get(
        self,
        count=None,
        as_string=False,
        as_numpy=True,
        timeout=None,
        with_ctrlvars=False,
        use_monitor=True,
    ):
        """Get value from MOCKPV"""
        data = self.get_with_metadata(
            count=count,
            as_string=as_string,
            as_numpy=as_numpy,
            timeout=timeout,
            with_ctrlvars=with_ctrlvars,
            use_monitor=use_monitor,
        )
        return data["value"] if data is not None else None


class DeviceMock:
    """Device Mock. Used for testing in combination with the DeviceManagerMock

    Args:
        name (str): name of the device
        value (float, optional): initial value of the device. Defaults to 0.0.
    Returns:
        DeviceMock: DeviceMock object

    """

    def __init__(self, name: str, value: float = 0.0):
        self.name = name
        self.read_buffer = value
        self._config = {"deviceConfig": {"limits": [-50, 50]}, "userParameter": None}
        self._read_only = False
        self._enabled = True

    def read(self):
        """Read method for DeviceMock"""
        return {self.name: {"value": self.read_buffer}}

    def readback(self):
        """Readback method for DeviceMock"""
        return self.read_buffer

    @property
    def read_only(self) -> bool:
        """read only property"""
        return self._read_only

    @read_only.setter
    def read_only(self, val: bool):
        """read only setter"""
        self._read_only = val

    @property
    def enabled(self) -> bool:
        """enabled property"""
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        """enabled setter"""
        self._enabled = val

    @property
    def user_parameter(self):
        """user_parameter property"""
        return self._config["userParameter"]

    @property
    def obj(self):
        """obj property"""
        return self


class DMMock:
    """Mock for DeviceManager

    The mocked DeviceManager creates a device containert and a connector.

    """

    def __init__(self):
        self.devices = DeviceContainer()
        self.connector = ConnectorMock()

    def add_device(self, name: str, value: float = 0.0):
        """Add device to the DeviceManagerMock"""
        self.devices[name] = DeviceMock(name, value)


# #TODO check what is the difference to SynSignal!
# class MockSignal(Signal):
#     """Can mock an OphydSignal"""
#     def __init__(self, read_pv, *, string=False, name=None, parent=None, **kwargs):
#         self.read_pv = read_pv
#         self._string = bool(string)
#         super().__init__(name=name, parent=parent, **kwargs)
#         self._waited_for_connection = False
#         self._subscriptions = []

#     def wait_for_connection(self):
#         self._waited_for_connection = True

#     def subscribe(self, method, event_type, **kw):
#         self._subscriptions.append((method, event_type, kw))

#     def describe_configuration(self):
#         return {self.name + "_conf": {"source": "SIM:test"}}

#     def read_configuration(self):
#         return {self.name + "_conf": {"value": 0}}



import enum
import os
import threading
import time
from typing import Any

import numpy as np
from bec_lib import messages, threadlocked
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from ophyd import ADComponent as ADCpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from std_daq_client import StdDaqClient

from ophyd_devices.epics.devices.psi_detector_base import CustomDetectorMixin, PSIDetectorBase

logger = bec_logger.logger


class EigerError(Exception):
    """Base class for exceptions in this module."""


class EigerTimeoutError(EigerError):
    """Raised when the Eiger does not respond in time."""


class Eiger9MSetup(CustomDetectorMixin):
    """Eiger setup class

    Parent class: CustomDetectorMixin

    """

    def __init__(self, *args, parent: Device = None, **kwargs) -> None:
        super().__init__(*args, parent=parent, **kwargs)
        self.std_rest_server_url = (
            kwargs["file_writer_url"] if "file_writer_url" in kwargs else "http://xbl-daq-29:5000"
        )
        self.std_client = None
        self._lock = threading.RLock()

    def initialize_default_parameter(self) -> None:
        """Set default parameters for Eiger9M detector"""
        self.update_readout_time()

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)

    def initialize_detector(self) -> None:
        """Initialize detector"""
        # Stops the detector
        self.stop_detector()
        # Sets the trigger source to GATING
        self.parent.set_trigger(TriggerSource.GATING)

    def initialize_detector_backend(self) -> None:
        """Initialize detector backend"""

        # Std client
        self.std_client = StdDaqClient(url_base=self.std_rest_server_url)

        # Stop writer
        self.std_client.stop_writer()

        # Change e-account
        eacc = self.parent.scaninfo.username
        self.update_std_cfg("writer_user_id", int(eacc.strip(" e")))

        signal_conditions = [(lambda: self.std_client.get_status()["state"], "READY")]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions, timeout=self.parent.timeout, all_signals=True
        ):
            raise EigerTimeoutError(
                f"Std client not in READY state, returns: {self.std_client.get_status()}"
            )

    def update_std_cfg(self, cfg_key: str, value: Any) -> None:
        """
        Update std_daq config

        Checks that the new value matches the type of the former entry.

        Args:
            cfg_key (str)   : config key of value to be updated
            value (Any)     : value to be updated for the specified key

        Raises:
            Raises EigerError if the key was not in the config before and if the new value does not match the type of the old value

        """

        # Load config from client and check old value
        cfg = self.std_client.get_config()
        old_value = cfg.get(cfg_key)
        if old_value is None:
            raise EigerError(
                f"Tried to change entry for key {cfg_key} in std_config that does not exist"
            )
        if not isinstance(value, type(old_value)):
            raise EigerError(
                f"Type of new value {type(value)}:{value} does not match old value"
                f" {type(old_value)}:{old_value}"
            )

        # Update config with new value and send back to client
        cfg.update({cfg_key: value})
        logger.debug(cfg)
        self.std_client.set_config(cfg)
        logger.debug(f"Updated std_daq config for key {cfg_key} from {old_value} to {value}")

    def stop_detector(self) -> None:
        """Stop the detector"""

        # Stop detector
        self.parent.cam.acquire.put(0)

        # Check if detector returned in idle state
        signal_conditions = [
            (
                lambda: self.parent.cam.detector_state.read()[self.parent.cam.detector_state.name][
                    "value"
                ],
                DetectorState.IDLE,
            )
        ]

        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout - self.parent.timeout // 2,
            check_stopped=True,
            all_signals=False,
        ):
            # Retry stop detector and wait for remaining time
            self.parent.cam.acquire.put(0)
            if not self.wait_for_signals(
                signal_conditions=signal_conditions,
                timeout=self.parent.timeout - self.parent.timeout // 2,
                check_stopped=True,
                all_signals=False,
            ):
                raise EigerTimeoutError(
                    f"Failed to stop detector, detector state {signal_conditions[0][0]}"
                )

    def stop_detector_backend(self) -> None:
        """Close file writer"""
        self.std_client.stop_writer()

    def prepare_detector(self) -> None:
        """Prepare detector for scan"""
        self.set_detector_threshold()
        self.set_acquisition_params()
        self.parent.set_trigger(TriggerSource.GATING)

    def set_detector_threshold(self) -> None:
        """
        Set the detector threshold

        The function sets the detector threshold automatically to 1/2 of the beam energy.
        """

        # get current beam energy from device manageer
        mokev = self.parent.device_manager.devices.mokev.obj.read()[
            self.parent.device_manager.devices.mokev.name
        ]["value"]
        factor = 1

        # Check if energies are eV or keV, assume keV as the default
        unit = getattr(self.parent.cam.threshold_energy, "units", None)
        if unit is not None and unit == "eV":
            factor = 1000

        # set energy on detector
        setpoint = int(mokev * factor)
        energy = self.parent.cam.beam_energy.read()[self.parent.cam.beam_energy.name]["value"]
        if setpoint != energy:
            self.parent.cam.beam_energy.set(setpoint)

        # set threshold on detector
        threshold = self.parent.cam.threshold_energy.read()[self.parent.cam.threshold_energy.name][
            "value"
        ]
        if not np.isclose(setpoint / 2, threshold, rtol=0.05):
            self.parent.cam.threshold_energy.set(setpoint / 2)

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for the detector"""

        # Set number of images and frames (frames is for internal burst of detector)
        self.parent.cam.num_images.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.cam.num_frames.put(1)

        # Update the readout time of the detector
        self.update_readout_time()

    def prepare_data_backend(self) -> None:
        """Prepare the data backend for the scan"""
        self.parent.filepath = self.parent.filewriter.compile_full_filename(
            f"{self.parent.name}.h5"
        )
        self.filepath_exists(self.parent.filepath)
        self.stop_detector_backend()
        try:
            self.std_client.start_writer_async(
                {
                    "output_file": self.parent.filepath,
                    "n_images": int(
                        self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                    ),
                }
            )
        except Exception as exc:
            time.sleep(5)
            if self.std_client.get_status()["state"] == "READY":
                raise EigerTimeoutError(f"Timeout of start_writer_async with {exc}") from exc

        # Check status of std_daq
        signal_conditions = [
            (lambda: self.std_client.get_status()["acquisition"]["state"], "WAITING_IMAGES")
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=False,
            all_signals=True,
        ):
            raise EigerTimeoutError(
                "Timeout of 5s reached for std_daq start_writer_async with std_daq client status"
                f" {self.std_client.get_status()}"
            )

    def filepath_exists(self, filepath: str) -> None:
        """Check if filepath exists"""
        signal_conditions = [(lambda: os.path.exists(os.path.dirname(filepath)), True)]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=False,
            all_signals=True,
        ):
            raise EigerError(f"Timeout of 3s reached for filepath {filepath}")

    def arm_acquisition(self) -> None:
        """Arm Eiger detector for acquisition"""
        self.parent.cam.acquire.put(1)
        signal_conditions = [
            (
                lambda: self.parent.cam.detector_state.read()[self.parent.cam.detector_state.name][
                    "value"
                ],
                DetectorState.RUNNING,
            )
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=False,
        ):
            raise EigerTimeoutError(
                f"Failed to arm the acquisition. Detector state {signal_conditions[0][0]}"
            )

    def check_scan_id(self) -> None:
        """Checks if scan_id has changed and stops the scan if it has"""
        old_scan_id = self.parent.scaninfo.scan_id
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scan_id != old_scan_id:
            self.parent.stopped = True

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the filepath to REDIS.

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
        """
        pipe = self.parent.connector.pipeline()
        if successful is None:
            msg = messages.FileMessage(file_path=self.parent.filepath, done=done)
        else:
            msg = messages.FileMessage(
                file_path=self.parent.filepath, done=done, successful=successful
            )
        self.parent.connector.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scan_id, self.parent.name),
            msg,
            pipe=pipe,
        )
        self.parent.connector.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg, pipe=pipe
        )
        pipe.execute()

    @threadlocked
    def finished(self):
        """Check if acquisition is finished."""
        signal_conditions = [
            (
                lambda: self.parent.cam.acquire.read()[self.parent.cam.acquire.name]["value"],
                DetectorState.IDLE,
            ),
            (lambda: self.std_client.get_status()["acquisition"]["state"], "FINISHED"),
            (
                lambda: self.std_client.get_status()["acquisition"]["stats"]["n_write_completed"],
                int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger),
            ),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            raise EigerTimeoutError(
                f"Reached timeout with detector state {signal_conditions[0][0]}, std_daq state"
                f" {signal_conditions[1][0]} and received frames of {signal_conditions[2][0]} for"
                " the file writer"
            )
        self.stop_detector()
        self.stop_detector_backend()


class SLSDetectorCam(Device):
    """
    SLS Detector Camera - Eiger9M

    Base class to map EPICS PVs to ophyd signals.
    """

    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    beam_energy = ADCpt(EpicsSignalWithRBV, "BeamEnergy")
    bit_depth = ADCpt(EpicsSignalWithRBV, "BitDepth")
    num_images = ADCpt(EpicsSignalWithRBV, "NumCycles")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumFrames")
    trigger_mode = ADCpt(EpicsSignalWithRBV, "TimingMode")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")
    acquire = ADCpt(EpicsSignal, "Acquire")
    detector_state = ADCpt(EpicsSignalRO, "DetectorState_RBV")


class TriggerSource(enum.IntEnum):
    """Trigger signals for Eiger9M detector"""

    AUTO = 0
    TRIGGER = 1
    GATING = 2
    BURST_TRIGGER = 3


class DetectorState(enum.IntEnum):
    """Detector states for Eiger9M detector"""

    IDLE = 0
    ERROR = 1
    WAITING = 2
    FINISHED = 3
    TRANSMITTING = 4
    RUNNING = 5
    STOPPED = 6
    STILL_WAITING = 7
    INITIALIZING = 8
    DISCONNECTED = 9
    ABORTED = 10


class Eiger9McSAXS(PSIDetectorBase):
    """
    Eiger9M detector for CSAXS

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (FalconSetup)        : Custom detector setup class for cSAXS,
                                                  inherits from CustomDetectorMixin
        PSIDetectorBase.set_min_readout (float) : Minimum readout time for the detector
        Various EpicsPVs for controlling the detector
    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe"]

    # specify Setup class
    custom_prepare_cls = Eiger9MSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3
    # specify class attributes
    cam = ADCpt(SLSDetectorCam, "cam1:")

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source for the detector.
        Check the TriggerSource enum for possible values

        Args:
            trigger_source (TriggerSource): Trigger source for the detector

        """
        value = trigger_source
        self.cam.trigger_mode.put(value)

    def stage(self) -> list[object]:
        """
        Add functionality to stage, and arm the detector

        Additional call to:
        - custom_prepare.arm_acquisition()
        """
        rtr = super().stage()
        self.custom_prepare.arm_acquisition()
        return rtr


if __name__ == "__main__":
    eiger = Eiger9McSAXS(name="eiger", prefix="X12SA-ES-EIGER9M:", sim_mode=True)



import pytest

from ophyd_devices.npoint import NPointAxis, NPointController


class SocketMock:
    def __init__(self, sock=None):
        self.buffer_put = ""
        self.buffer_recv = ""
        self.is_open = False
        if sock is None:
            self.open()
        else:
            self.sock = sock

    def connect(self, host, port):
        print(f"connecting to {host} port {port}")
        # self.sock.create_connection((host, port))
        # self.sock.connect((host, port))

    def _put(self, msg_bytes):
        self.buffer_put = msg_bytes
        print(self.buffer_put)

    def _recv(self, buffer_length=1024):
        print(self.buffer_recv)
        return self.buffer_recv

    def _initialize_socket(self):
        pass

    def put(self, msg):
        return self._put(msg)

    def receive(self, buffer_length=1024):
        return self._recv(buffer_length=buffer_length)

    def open(self):
        self._initialize_socket()
        self.is_open = True

    def close(self):
        self.sock = None
        self.is_open = False


@pytest.mark.parametrize(
    "pos,msg",
    [
        (5, b"\xa2\x18\x12\x83\x11\xcd\xcc\x00\x00U"),
        (0, b"\xa2\x18\x12\x83\x11\x00\x00\x00\x00U"),
        (-5, b"\xa2\x18\x12\x83\x1133\xff\xffU"),
    ],
)
def test_axis_put(pos, msg):
    controller = NPointController(SocketMock())
    npointx = NPointAxis(controller, 0, "nx")
    controller.on()
    npointx.set(pos)
    assert npointx.controller.socket.buffer_put == msg


@pytest.mark.parametrize(
    "pos, msg_in, msg_out",
    [
        (5.0, b"\xa04\x13\x83\x11U", b"\xa0\x34\x13\x83\x11\xcd\xcc\x00\x00U"),
        (0, b"\xa04\x13\x83\x11U", b"\xa0\x34\x13\x83\x11\x00\x00\x00\x00U"),
        (-5, b"\xa04\x13\x83\x11U", b"\xa0\x34\x13\x83\x1133\xff\xffU"),
    ],
)
def test_axis_get_out(pos, msg_in, msg_out):
    controller = NPointController(SocketMock())
    npointx = NPointAxis(controller, 0, "nx")
    controller.on()
    npointx.controller.socket.buffer_recv = msg_out
    assert pytest.approx(npointx.get(), rel=0.01) == pos
    # assert controller.socket.buffer_put == msg_in


@pytest.mark.parametrize(
    "axis, msg_in, msg_out",
    [
        (0, b"\xa04\x13\x83\x11U", b"\xa0\x34\x13\x83\x11\xcd\xcc\x00\x00U"),
        (1, b"\xa04#\x83\x11U", b"\xa0\x34\x13\x83\x11\x00\x00\x00\x00U"),
        (2, b"\xa043\x83\x11U", b"\xa0\x34\x13\x83\x1133\xff\xffU"),
    ],
)
def test_axis_get_in(axis, msg_in, msg_out):
    controller = NPointController(SocketMock())
    npointx = NPointAxis(controller, 0, "nx")
    controller.on()
    controller.socket.buffer_recv = msg_out
    controller._get_current_pos(axis)
    assert controller.socket.buffer_put == msg_in


def test_axis_out_of_range():
    controller = NPointController(SocketMock())
    with pytest.raises(ValueError):
        npointx = NPointAxis(controller, 3, "nx")


def test_get_axis_out_of_range():
    controller = NPointController(SocketMock())
    with pytest.raises(ValueError):
        controller._get_current_pos(3)


def test_set_axis_out_of_range():
    controller = NPointController(SocketMock())
    with pytest.raises(ValueError):
        controller._set_target_pos(3, 5)


@pytest.mark.parametrize(
    "in_buffer, byteorder, signed, val",
    [
        (["0x0", "0x0", "0xcc", "0xcd"], "big", True, 52429),
        (["0xcd", "0xcc", "0x0", "0x0"], "little", True, 52429),
        (["cd", "cc", "00", "00"], "little", True, 52429),
    ],
)
def test_hex_list_to_int(in_buffer, byteorder, signed, val):
    assert NPointController._hex_list_to_int(in_buffer, byteorder=byteorder, signed=signed) == val


@pytest.mark.parametrize(
    "axis, msg_in, msg_out",
    [
        (0, b"\xa0x\x10\x83\x11U", b"\xa0\x78\x13\x83\x11\x64\x00\x00\x00U"),
        (1, b"\xa0x \x83\x11U", b"\xa0\x78\x13\x83\x11\x64\x00\x00\x00U"),
        (2, b"\xa0x0\x83\x11U", b"\xa0\x78\x13\x83\x11\x64\x00\x00\x00U"),
    ],
)
def test_get_range(axis, msg_in, msg_out):
    controller = NPointController(SocketMock())
    npointx = NPointAxis(controller, 0, "nx")
    controller.on()
    controller.socket.buffer_recv = msg_out
    val = controller._get_range(axis)
    assert controller.socket.buffer_put == msg_in and val == 100



from .bec_protocols import (
    BECDeviceProtocol,
    BECFlyerProtocol,
    BECMixinProtocol,
    BECPositionerProtocol,
    BECRotationProtocol,
    BECScanProtocol,
    BECSignalProtocol,
)
from .tomcat_rotation_motors import TomcatAerotechRotation



import socket

from ophyd_devices.utils.socket import SocketIO


class DummySocket:
    AF_INET = 2
    SOCK_STREAM = 1

    def __init__(self) -> None:
        self.address_family = None
        self.socket_kind = None
        self.timeout = None

    def socket(self, address_family, socket_kind):
        self.address_family = address_family
        self.socket_kind = socket_kind
        return self

    def settimeout(self, timeout):
        self.timeout = timeout

    def send(self, msg, *args, **kwargs):
        self.send_buffer = msg

    def connect(self, address):
        self.host = address[0]
        self.port = address[1]
        self.connected = True

    def close(self):
        self.connected = False


def test_socket_init():
    socketio = SocketIO("localhost", 8080)

    assert socketio.host == "localhost"
    assert socketio.port == 8080

    assert socketio.is_open == False

    assert socketio.sock.family == socket.AF_INET
    assert socketio.sock.type == socket.SOCK_STREAM


def test_socket_put():
    dsocket = DummySocket()
    socketio = SocketIO("localhost", 8080)
    socketio.sock = dsocket
    socketio.put(b"message")
    assert dsocket.send_buffer == b"message"


def test_open():
    dsocket = DummySocket()
    socketio = SocketIO("localhost", 8080)
    socketio.sock = dsocket
    socketio.open()
    assert socketio.is_open == True
    assert socketio.sock.host == socketio.host
    assert socketio.sock.port == socketio.port


def test_close():
    socketio = SocketIO("localhost", 8080)
    socketio.close()
    assert socketio.sock == None
    assert socketio.is_open == False



import threading
import time

import numpy as np
from bec_lib import MessageEndpoints, messages
from ophyd import Component as Cpt
from ophyd import Device, Kind, Signal
from ophyd.flyers import FlyerInterface
from ophyd.ophydobj import Kind
from ophyd.status import DeviceStatus, SubscriptionStatus
from ophyd.utils import ReadOnlyError


class SynSetpoint(Signal):
    def __init__(
        self,
        name,
        *,
        value=0,
        dtype=float,
        timestamp=None,
        parent=None,
        labels=None,
        kind=Kind.hinted,
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        auto_monitor=False,
    ):
        self._dtype = dtype
        if self._dtype == str and value == 0:
            value = ""
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )

    def put(self, value, *, timestamp=None, force=False):
        old_val = self._readback
        self._readback = self._dtype(value)
        self._run_subs(
            sub_type="value", old_value=old_val, value=self._readback, timestamp=time.time()
        )

    def get(self):
        return self._readback

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res


class SynData(Signal):
    _default_sub = "value"

    def __init__(
        self,
        *,
        name,
        value=0,
        timestamp=None,
        parent=None,
        labels=None,
        kind=Kind.hinted,
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        auto_monitor=False,
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self._reset_data()

    def _reset_data(self):
        self._readback = np.array([])

    def get(self):
        return self._readback

    def append(self, val: float):
        self._readback = np.append(self._readback, val)

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of
        # generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return time.time()

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class SynXtremeOtf(FlyerInterface, Device):
    """
    PGM on-the-fly scan
    """

    SUB_VALUE = "value"
    SUB_FLYER = "flyer"
    _default_sub = SUB_VALUE

    e1 = Cpt(SynSetpoint, kind=Kind.config)
    e2 = Cpt(SynSetpoint, kind=Kind.config)
    time = Cpt(SynSetpoint, kind=Kind.config)
    folder = Cpt(SynSetpoint, dtype=str, value="", kind=Kind.config)
    file = Cpt(SynSetpoint, dtype=str, value="", kind=Kind.config)
    acquire = Cpt(SynSetpoint, auto_monitor=True)
    edata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    data = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    idata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    fdata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    count = Cpt(SynData, kind=Kind.omitted, auto_monitor=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_time = 0
        self.acquire.subscribe(self._update_status, run=False)
        self.count.subscribe(self._update_data, run=False)
        self._data_event = threading.Event()
        self.precision = 3

    def kickoff(self):
        self._start_time = time.time()
        self.acquire.put(1)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def complete(self):
        def check_value(*, old_value, value, **kwargs):
            return old_value == 1 and value == 0

        status = SubscriptionStatus(self.acquire, check_value, event_type=self.acquire.SUB_VALUE)
        return status

    def collect(self):
        data = {"time": self._start_time, "data": {}, "timestamps": {}}
        for attr in ("edata", "data", "idata", "fdata"):
            obj = getattr(self, attr)
            data["data"][obj.name] = obj.get()
            data["timestamps"][obj.name] = obj.timestamp

        return data

    def describe_collect(self):
        desc = {}
        for attr in ("edata", "data", "idata", "fdata"):
            desc.update(getattr(self, attr).describe())
        return desc

    def _update_status(self, *, old_value, value, **kwargs):
        if old_value == 1 and value == 0:
            self._done_acquiring()
            return
        if old_value == 0 and value == 1:
            threading.Thread(target=self._start_acquiring, daemon=True).start()

    def _reset_data(self):
        for entry in ("edata", "data", "idata", "fdata"):
            getattr(self, entry)._reset_data()
        self.count._readback = 0
        self._data_event.clear()

    def _populate_data(self):
        self._reset_data()
        while not self._data_event.is_set():
            for entry in ("edata", "data", "idata", "fdata"):
                getattr(self, entry).append(np.random.rand())
            self.count._readback = len(self.edata.get())
            self.count._run_subs(
                sub_type="value",
                old_value=self.count._readback - 1,
                value=self.count._readback,
                timestamp=time.time(),
            )
            time.sleep(0.2)
        self._data_event.clear()

    def _start_acquiring(self):
        threading.Thread(target=self._populate_data, daemon=True).start()
        timeout_event = threading.Event()
        flag = timeout_event.wait(self.time.get())
        if not flag:
            self._data_event.set()
            self.acquire.put(0)

    def _update_data(self, value, **kwargs):
        if value == 0:
            return
        data = self.collect()
        self._run_subs(sub_type=self.SUB_FLYER, value=data)


class SynXtremeOtfReplay(FlyerInterface, Device):
    """
    PGM on-the-fly scan
    """

    SUB_VALUE = "value"
    SUB_FLYER = "flyer"
    _default_sub = SUB_VALUE

    e1 = Cpt(SynSetpoint, kind=Kind.config)
    e2 = Cpt(SynSetpoint, kind=Kind.config)
    time = Cpt(SynSetpoint, kind=Kind.config)
    folder = Cpt(SynSetpoint, dtype=str, value="", kind=Kind.config)
    file = Cpt(SynSetpoint, dtype=str, value="", kind=Kind.config)
    acquire = Cpt(SynSetpoint, auto_monitor=True)
    edata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    data = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    idata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    fdata = Cpt(SynData, kind=Kind.hinted, auto_monitor=True)
    count = Cpt(SynData, kind=Kind.omitted, auto_monitor=True)

    def __init__(self, *args, device_manager=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_time = 0
        self.acquire.subscribe(self._update_status, run=False)
        self.count.subscribe(self._update_data, run=False)
        self._data_event = threading.Event()
        self.precision = 3
        self._measurement_data = {}
        self._device_manager = device_manager

    def kickoff(self):
        self._read_file()
        self._start_time = time.time()
        self.acquire.put(1)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def _read_file(self):
        data = {}  # Create an empty dictionary to store the data

        # Read the input file
        with open(self.file.get(), "r") as file:
            lines = file.readlines()

        # Extract column titles and data types
        titles = lines[0].strip().split("\t")
        data_types = lines[1].strip().split("\t")

        # Initialize the dictionary with empty lists for each title
        for title in titles:
            data[title] = []

        # Process the data lines and populate the dictionary
        for line in lines[2:]:
            values = line.strip().split("\t")
            for title, value, data_type in zip(titles, values, data_types):
                if data_type == "java.lang.Double":
                    data[title].append(float(value))
                else:
                    data[title].append(value)
        self._measurement_data = data

    def complete(self):
        def check_value(*, old_value, value, **kwargs):
            return old_value == 1 and value == 0

        if self.acquire.get() == 0:
            status = DeviceStatus(self)
            status.set_finished()
            return status
        status = SubscriptionStatus(self.acquire, check_value, event_type=self.acquire.SUB_VALUE)
        return status

    def collect(self):
        self._update_measurement_data()
        data = {"time": self._start_time, "data": {}, "timestamps": {}}
        for attr in ("edata", "data", "idata", "fdata"):
            obj = getattr(self, attr)
            if attr == "edata":
                data["data"][obj.name] = np.asarray(
                    [float(x) for x in self._measurement_data["#Ecrbk"][7 : 7 + self.count.get()]]
                )
            else:
                data["data"][obj.name] = obj.get()
            data["timestamps"][obj.name] = obj.timestamp

        return data

    def _update_measurement_data(self):
        timestamp = time.time()
        signals = {
            "signals_s1": {
                "value": self._measurement_data["CADC1"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_s2": {
                "value": self._measurement_data["CADC2"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_s3": {
                "value": self._measurement_data["CADC3"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_s4": {
                "value": self._measurement_data["CADC4"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_s5": {
                "value": self._measurement_data["CADC5"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_norm_tey": {
                "value": self._measurement_data["CADC1"][self.count.get()]
                / self._measurement_data["CADC2"][self.count.get()],
                "timestamp": timestamp,
            },
            "signals_norm_diode": {
                "value": self._measurement_data["CADC1"][self.count.get()]
                / self._measurement_data["CADC3"][self.count.get()],
                "timestamp": timestamp,
            },
        }
        msg = messages.DeviceMessage(
            signals=signals, metadata=self._device_manager.devices.otf.metadata
        )
        self._device_manager.connector.set_and_publish(
            MessageEndpoints.device_readback("signals"), msg
        )

    def describe_collect(self):
        desc = {}
        for attr in ("edata", "data", "idata", "fdata"):
            desc.update(getattr(self, attr).describe())
        return desc

    def _update_status(self, *, old_value, value, **kwargs):
        if old_value == 1 and value == 0:
            self._done_acquiring()
            return
        if old_value == 0 and value == 1:
            threading.Thread(target=self._start_acquiring, daemon=True).start()

    def _reset_data(self):
        for entry in ("edata", "data", "idata", "fdata"):
            getattr(self, entry)._reset_data()
        self.count._readback = 0
        self._data_event.clear()

    def _populate_data(self):
        self._reset_data()
        while not self._data_event.is_set():
            for entry in ("edata", "data", "idata", "fdata"):
                getattr(self, entry).append(np.random.rand())
            self.count._readback = len(self.edata.get())
            self.count._run_subs(
                sub_type="value",
                old_value=self.count._readback - 1,
                value=self.count._readback,
                timestamp=time.time(),
            )
            time.sleep(0.1)
        self._data_event.clear()

    def _start_acquiring(self):
        threading.Thread(target=self._populate_data, daemon=True).start()
        timeout_event = threading.Event()
        flag = timeout_event.wait(self.time.get())
        if not flag:
            self._data_event.set()
            self.acquire.put(0)

    def _update_data(self, value, **kwargs):
        if value == 0:
            return
        data = self.collect()
        self._run_subs(sub_type=self.SUB_FLYER, value=data)


if __name__ == "__main__":
    obj = SynXtremeOtf(name="otf")
    status = obj.time.set(4)
    status.wait()
    status = obj.kickoff()
    status.wait()
    while obj.acquire.get():
        time.sleep(0.2)
    print("done")



import time

from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import EpicsSignal
from prettytable import PrettyTable


class FlomniSampleStorageError(Exception):
    pass


class FlomniSampleStorage(Device):
    USER_ACCESS = [
        "is_sample_slot_used",
        "is_sample_in_gripper",
        "set_sample_slot",
        "unset_sample_slot",
        "set_sample_in_gripper",
        "unset_sample_in_gripper",
        "show_all",
    ]
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE
    sample_placed = {
        f"sample{i}": (EpicsSignal, f"XOMNY-SAMPLE_DB_flomni{i}:GET", {}) for i in range(21)
    }
    sample_placed = Dcpt(sample_placed)

    sample_names = {
        f"sample{i}": (EpicsSignal, f"XOMNY-SAMPLE_DB_flomni{i}:GET.DESC", {"string": True})
        for i in range(21)
    }
    sample_names = Dcpt(sample_names)

    sample_in_gripper = Cpt(
        EpicsSignal, name="sample_in_gripper", read_pv="XOMNY-SAMPLE_DB_flomni100:GET"
    )
    sample_in_gripper_name = Cpt(
        EpicsSignal,
        name="sample_in_gripper_name",
        read_pv="XOMNY-SAMPLE_DB_flomni100:GET.DESC",
        string=True,
    )

    def __init__(self, prefix="", *, name, **kwargs):
        super().__init__(prefix, name=name, **kwargs)
        self.sample_placed.sample1.subscribe(self._emit_value)

    def _emit_value(self, **kwargs):
        timestamp = kwargs.pop("timestamp", time.time())
        self.wait_for_connection()
        self._run_subs(sub_type=self.SUB_VALUE, timestamp=timestamp, obj=self)

    def set_sample_slot(self, slot_nr: int, name: str) -> bool:
        if slot_nr > 20:
            raise FlomniSampleStorageError(f"Invalid slot number {slot_nr}.")

        getattr(self.sample_placed, f"sample{slot_nr}").set(1)
        getattr(self.sample_names, f"sample{slot_nr}").set(name)

    def unset_sample_slot(self, slot_nr: int) -> bool:
        if slot_nr > 20:
            raise FlomniSampleStorageError(f"Invalid slot number {slot_nr}.")

        getattr(self.sample_placed, f"sample{slot_nr}").set(0)
        getattr(self.sample_names, f"sample{slot_nr}").set("-")

    def set_sample_in_gripper(self, name: str) -> bool:
        self.sample_in_gripper.set(1)
        self.sample_in_gripper_name.set(name)

    def unset_sample_in_gripper(self) -> bool:
        self.sample_in_gripper.set(0)
        self.sample_in_gripper_name.set("-")

    def is_sample_slot_used(self, slot_nr: int) -> bool:
        val = getattr(self.sample_placed, f"sample{slot_nr}").get()
        return bool(val)

    def is_sample_in_gripper(self) -> bool:
        val = self.sample_in_gripper.get()
        return bool(val)

    def get_sample_name(self, slot_nr) -> str:
        val = getattr(self.sample_names, f"sample{slot_nr}").get()
        return str(val)

    def show_all(self):
        t = PrettyTable()
        t.title = "flOMNI sample storage"
        field_names = [""]
        field_names.extend(str(ax) for ax in range(1, 11))
        for ct in range(0, 2):
            t.field_names = field_names
            row = ["Container " + str(ct)]
            row.extend(
                "used" if self.is_sample_slot_used(slot_nr) else "free"
                for slot_nr in range((ct * 10) + 1, (ct * 10) + 11)
            )
            t.add_row(row)
        print(t)
        print("\n\nFollowing samples are currently loaded:\n")
        for ct in range(1, 21):
            if self.is_sample_slot_used(ct):
                print(f"  Position {ct:2.0f}: {self.get_sample_name(ct)}")
        if self.sample_in_gripper.get():
            print(f"\n  Gripper:        {self.sample_in_gripper_name.get()}\n")
        else:
            print(f"\n  Gripper:        no sample\n")

        if self.is_sample_slot_used(0):
            print(f"  flOMNI stage:   {self.get_sample_name(0)}\n")
        else:
            print(f"  flOMNI stage:   no sample\n")



import time
from collections import OrderedDict
from time import sleep

import numpy as np
from ophyd import Component, Device, EpicsMotor, EpicsSignal, EpicsSignalRO, Kind
from ophyd.status import DeviceStatus, Status, StatusBase, SubscriptionStatus

from ophyd_devices.epics.devices.aerotech.AerotechAutomation1Enums import (
    DataCollectionFrequency,
    DataCollectionMode,
    DriveDataCaptureInput,
    DriveDataCaptureTrigger,
)


class EpicsMotorX(EpicsMotor):
    """Special motor class that provides flyer interface and progress bar."""

    SUB_PROGRESS = "progress"

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self._startPosition = None
        self._targetPosition = None
        self.subscribe(self._progress_update, run=False)

    def configure(self, d: dict):
        if "target" in d:
            self._targetPosition = d["target"]
            del d["target"]
        if "position" in d:
            self._targetPosition = d["position"]
            del d["position"]
        return super().configure(d)

    def kickoff(self):
        self._startPosition = float(self.position)
        return self.move(self._targetPosition, wait=False)

    def move(self, position, wait=True, **kwargs):
        self._startPosition = float(self.position)
        return super().move(position, wait, **kwargs)

    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan"""
        if (self._startPosition is None) or (self._targetPosition is None) or (not self.moving):
            self._run_subs(sub_type=self.SUB_PROGRESS, value=1, max_value=1, done=1)
            return

        progress = np.abs(
            (value - self._startPosition) / (self._targetPosition - self._startPosition)
        )
        max_value = 100
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=int(100 * progress),
            max_value=max_value,
            done=int(np.isclose(max_value, progress, 1e-3)),
        )


class EpicsPassiveRO(EpicsSignalRO):
    """Small helper class to read PVs that need to be processed first."""

    def __init__(self, read_pv, *, string=False, name=None, **kwargs):
        super().__init__(read_pv, string=string, name=name, **kwargs)
        self._proc = EpicsSignal(read_pv + ".PROC", kind=Kind.omitted, put_complete=True)

    def wait_for_connection(self, *args, **kwargs):
        super().wait_for_connection(*args, **kwargs)
        self._proc.wait_for_connection(*args, **kwargs)

    def get(self, *args, **kwargs):
        self._proc.set(1).wait()
        return super().get(*args, **kwargs)

    @property
    def value(self):
        return super().value


class aa1Controller(Device):
    """Ophyd proxy class for the Aerotech Automation 1's core controller functionality"""

    controllername = Component(EpicsSignalRO, "NAME", kind=Kind.config)
    serialnumber = Component(EpicsSignalRO, "SN", kind=Kind.config)
    apiversion = Component(EpicsSignalRO, "API_VERSION", kind=Kind.config)
    axiscount = Component(EpicsSignalRO, "AXISCOUNT", kind=Kind.config)
    taskcount = Component(EpicsSignalRO, "TASKCOUNT", kind=Kind.config)
    fastpoll = Component(EpicsSignalRO, "POLLTIME", auto_monitor=True, kind=Kind.hinted)
    slowpoll = Component(EpicsSignalRO, "DRVPOLLTIME", auto_monitor=True, kind=Kind.hinted)

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )


class aa1Tasks(Device):
    """Task management API

    The place to manage tasks and AeroScript user files on the controller.
    You can read/write/compile/execute AeroScript files and also retrieve
    saved data files from the controller. It will also work around an ophyd
    bug that swallows failures.

    Execution does not require to store the script in a file, it will compile
    it and run it directly on a certain thread. But there's no way to
    retrieve the source code.

    Write a text into a file on the aerotech controller and execute it with kickoff.
    '''
    script="var $axis as axis = ROTY\\nMoveAbsolute($axis, 42, 90)"
    tsk = aa1Tasks(AA1_IOC_NAME+":TASK:", name="tsk")
    tsk.wait_for_connection()
    tsk.configure({'text': script, 'filename': "foobar.ascript", 'taskIndex': 4})
    tsk.kickoff().wait()
    '''

    Just execute an ascript file already on the aerotech controller.
    '''
    tsk = aa1Tasks(AA1_IOC_NAME+":TASK:", name="tsk")
    tsk.wait_for_connection()
    tsk.configure({'filename': "foobar.ascript", 'taskIndex': 4})
    tsk.kickoff().wait()
    '''

    """

    SUB_PROGRESS = "progress"
    _failure = Component(EpicsSignalRO, "FAILURE", auto_monitor=True, kind=Kind.hinted)
    errStatus = Component(EpicsSignalRO, "ERRW", auto_monitor=True, kind=Kind.hinted)
    warnStatus = Component(EpicsSignalRO, "WARNW", auto_monitor=True, kind=Kind.hinted)
    taskStates = Component(EpicsSignalRO, "STATES-RBV", auto_monitor=True, kind=Kind.hinted)
    taskIndex = Component(EpicsSignal, "TASKIDX", kind=Kind.config, put_complete=True)
    switch = Component(EpicsSignal, "SWITCH", kind=Kind.config, put_complete=True)
    _execute = Component(EpicsSignal, "EXECUTE", string=True, kind=Kind.config, put_complete=True)
    _executeMode = Component(EpicsSignal, "EXECUTE-MODE", kind=Kind.config, put_complete=True)
    _executeReply = Component(EpicsSignalRO, "EXECUTE_RBV", string=True, auto_monitor=True)

    fileName = Component(EpicsSignal, "FILENAME", string=True, kind=Kind.omitted, put_complete=True)
    _fileRead = Component(EpicsPassiveRO, "FILEREAD", string=True, kind=Kind.omitted)
    _fileWrite = Component(
        EpicsSignal, "FILEWRITE", string=True, kind=Kind.omitted, put_complete=True
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        """__init__ MUST have a full argument list"""
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self._currentTask = None
        self._textToExecute = None
        self._isConfigured = False
        self._isStepConfig = False
        self.subscribe(self._progress_update, "progress", run=False)

    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan"""
        value = self.progress()
        self._run_subs(sub_type=self.SUB_PROGRESS, value=value, max_value=1, done=1)

    def _progress(self) -> None:
        """Progress update on the scan"""
        if self._currentTaskMonitor is None:
            return 1
        else:
            if self._currentTaskMonitor.status.value in ["Running", 4]:
                return 0
            else:
                return 1

    def readFile(self, filename: str) -> str:
        """Read a file from the controller"""
        # Have to use CHAR array due to EPICS LSI bug...
        self.fileName.set(filename).wait()
        filebytes = self._fileRead.get()
        # C-strings terminate with trailing zero
        if filebytes[-1] == 0:
            filebytes = filebytes[:-1]
        filetext = filebytes
        return filetext

    def writeFile(self, filename: str, filetext: str) -> None:
        """Write a file to the controller"""
        self.fileName.set(filename).wait()
        self._fileWrite.set(filetext).wait()

    def runScript(self, filename: str, taskIndex: int == 2, filetext=None, settle_time=0.5) -> None:
        """Run a script file that either exists, or is newly created and compiled"""

        self.configure({"text": filetext, "filename": filename, "taskIndex": taskIndex})
        print("Runscript configured")
        self.trigger().wait()
        print("Runscript waited")

    def execute(self, text: str, taskIndex: int = 3, mode: str = 0, settle_time=0.5):
        """Run a short text command on the Automation1 controller"""

        print(f"Executing program on task: {taskIndex}")
        self.configure({"text": text, "taskIndex": taskIndex, "mode": mode})
        self.kickoff().wait()

        if mode in [0, "None", None]:
            return None
        else:
            raw = self._executeReply.get()
            return raw

    def configure(self, d: dict = {}) -> tuple:
        """Configuration interface for flying"""
        # Unrolling the configuration dict
        text = str(d["text"]) if "text" in d else None
        filename = str(d["filename"]) if "filename" in d else None
        taskIndex = int(d["taskIndex"]) if "taskIndex" in d else 4
        settle_time = float(d["settle_time"]) if "settle_time" in d else None
        mode = d["mode"] if "mode" in d else None
        self._isStepConfig = d["stepper"] if "stepper" in d else False

        # Validation
        if taskIndex < 1 or taskIndex > 31:
            raise RuntimeError(f"Invalid task index: {taskIndex}")
        if (text is None) and (filename is None):
            raise RuntimeError("Task execution requires either AeroScript text or filename")

        # Common operations
        old = self.read_configuration()
        self.taskIndex.set(taskIndex).wait()
        self._textToExecute = None
        self._currentTask = taskIndex

        # Choose the right execution mode
        if (filename is None) and (text not in [None, ""]):
            # Direct command execution
            print("Preparing for direct command execution")
            if mode is not None:
                self._executeMode.set(mode).wait()
            self._textToExecute = text
        elif (filename is not None) and (text in [None, ""]):
            # Execute existing file
            self.fileName.set(filename).wait()
            self.switch.set("Load").wait()
        elif (filename is not None) and (text not in [None, ""]):
            print("Preparing to execute via intermediary file")
            # Execute text via intermediate file
            self.taskIndex.set(taskIndex).wait()
            self.fileName.set(filename).wait()
            self._fileWrite.set(text).wait()
            self.switch.set("Load").wait()
            self._textToExecute = None
        else:
            raise RuntimeError("Unsupported filename-text combo")

        if self._failure.value:
            raise RuntimeError("Failed to launch task, please check the Aerotech IOC")

        self._isConfigured = True
        new = self.read_configuration()
        return (old, new)

    ##########################################################################
    # Bluesky stepper interface
    def stage(self) -> None:
        """Default staging"""
        super().stage()

    def unstage(self) -> None:
        """Default unstaging"""
        super().unstage()

    def trigger(self, settle_time=0.2) -> Status:
        """Execute the script on the configured task"""
        if self._isStepConfig:
            return self.kickoff(settle_time)
        else:
            status = DeviceStatus(self, settle_time=settle_time)
            status.set_finished()
            if settle_time is not None:
                sleep(settle_time)
        return status

    def stop(self):
        """Stop the currently selected task"""
        self.switch.set("Stop").wait()

    ##########################################################################
    # Flyer interface
    def kickoff(self, settle_time=0.2) -> DeviceStatus:
        """Execute the script on the configured task"""
        if self._isConfigured:
            if self._textToExecute is not None:
                print(f"Kickoff directly executing string: {self._textToExecute}")
                status = self._execute.set(self._textToExecute, settle_time=0.5)
            else:
                status = self.switch.set("Run", settle_time=0.1)
        else:
            status = DeviceStatus(self)
            status.set_finished()
        if settle_time is not None:
            sleep(settle_time)
        return status

    def complete(self) -> DeviceStatus:
        """Execute the script on the configured task"""
        print("Called aa1Task.complete()")
        timestamp_ = 0
        taskIdx = int(self.taskIndex.get())

        def notRunning2(*args, old_value, value, timestamp, **kwargs):
            nonlocal timestamp_
            result = False if value[taskIdx] in ["Running", 4] else True
            timestamp_ = timestamp
            print(result)
            return result

        # Subscribe and wait for update
        status = SubscriptionStatus(self.taskStates, notRunning2, settle_time=0.5)
        return status

    def describe_collect(self) -> OrderedDict:
        dd = OrderedDict()
        dd["success"] = {
            "source": "internal",
            "dtype": "integer",
            "shape": [],
            "units": "",
            "lower_ctrl_limit": 0,
            "upper_ctrl_limit": 0,
        }
        return {self.name: dd}

    def collect(self) -> OrderedDict:
        ret = OrderedDict()
        ret["timestamps"] = {"success": time.time()}
        ret["data"] = {"success": 1}
        yield ret


class aa1TaskState(Device):
    """Task state monitoring API

    This is the task state monitoring interface for Automation1 tasks. It
    does not launch execution, but can wait for the execution to complete.
    """

    index = Component(EpicsSignalRO, "INDEX", kind=Kind.config)
    status = Component(EpicsSignalRO, "STATUS", auto_monitor=True, kind=Kind.hinted)
    errorCode = Component(EpicsSignalRO, "ERROR", auto_monitor=True, kind=Kind.hinted)
    warnCode = Component(EpicsSignalRO, "WARNING", auto_monitor=True, kind=Kind.hinted)

    def complete(self) -> StatusBase:
        """Bluesky flyer interface"""
        print("Called aa1TaskState.complete()")
        # Define wait until the busy flag goes down (excluding initial update)
        timestamp_ = 0

        def notRunning(*args, old_value, value, timestamp, **kwargs):
            nonlocal timestamp_
            result = False if (timestamp_ == 0) else (value not in ["Running", 4])
            timestamp_ = timestamp
            print(result)
            return result

        # Subscribe and wait for update
        status = SubscriptionStatus(self.status, notRunning, settle_time=0.5)
        return status

    def kickoff(self) -> DeviceStatus:
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def describe_collect(self) -> OrderedDict:
        dd = OrderedDict()
        dd["success"] = {
            "source": "internal",
            "dtype": "integer",
            "shape": [],
            "units": "",
            "lower_ctrl_limit": 0,
            "upper_ctrl_limit": 0,
        }
        return dd

    def collect(self) -> OrderedDict:
        ret = OrderedDict()
        ret["timestamps"] = {"success": time.time()}
        ret["data"] = {"success": 1}
        yield ret


class aa1DataAcquisition(Device):
    """Controller Data Acquisition - DONT USE at Tomcat

    This class implements the controller data collection feature of the
    Automation1 controller. This feature logs various inputs at a
    **fixed frequency** from 1 kHz up to 200 kHz.
    Usage:
        1. Start a new configuration with "startConfig"
        2. Add your signals with "addXxxSignal"
        3. Start your data collection
        4. Read back the recorded data with "readback"
    """

    # Status monitoring
    status = Component(EpicsSignalRO, "RUNNING", auto_monitor=True, kind=Kind.hinted)
    points_max = Component(EpicsSignal, "MAXPOINTS", kind=Kind.config, put_complete=True)
    signal_num = Component(EpicsSignalRO, "NITEMS", kind=Kind.config)

    points_total = Component(EpicsSignalRO, "NTOTAL", auto_monitor=True, kind=Kind.hinted)
    points_collected = Component(EpicsSignalRO, "NCOLLECTED", auto_monitor=True, kind=Kind.hinted)
    points_retrieved = Component(EpicsSignalRO, "NRETRIEVED", auto_monitor=True, kind=Kind.hinted)
    overflow = Component(EpicsSignalRO, "OVERFLOW", auto_monitor=True, kind=Kind.hinted)
    runmode = Component(EpicsSignalRO, "MODE_RBV", auto_monitor=True, kind=Kind.hinted)
    # DAQ setup
    numpoints = Component(EpicsSignal, "NPOINTS", kind=Kind.config, put_complete=True)
    frequency = Component(EpicsSignal, "FREQUENCY", kind=Kind.config, put_complete=True)
    _configure = Component(EpicsSignal, "CONFIGURE", kind=Kind.omitted, put_complete=True)

    def startConfig(self, npoints: int, frequency: DataCollectionFrequency):
        self.numpoints.set(npoints).wait()
        self.frequency.set(frequency).wait()
        self._configure.set("START").wait()

    def clearConfig(self):
        self._configure.set("CLEAR").wait()

    srcTask = Component(EpicsSignal, "SRC_TASK", kind=Kind.config, put_complete=True)
    srcAxis = Component(EpicsSignal, "SRC_AXIS", kind=Kind.config, put_complete=True)
    srcCode = Component(EpicsSignal, "SRC_CODE", kind=Kind.config, put_complete=True)
    _srcAdd = Component(EpicsSignal, "SRC_ADD", kind=Kind.omitted, put_complete=True)

    def addAxisSignal(self, axis: int, code: int) -> None:
        """Add a new axis-specific data signal to the DAQ configuration. The
        most common signals are PositionFeedback and PositionError.
        """
        self.srcAxis.set(axis).wait()
        self.srcCode.set(code).wait()
        self._srcAdd.set("AXIS").wait()

    def addTaskSignal(self, task: int, code: int) -> None:
        """Add a new task-specific data signal to the DAQ configuration"""
        self.srcTask.set(task).wait()
        self.srcCode.set(code).wait()
        self._srcAdd.set("TASK").wait()

    def addSystemSignal(self, code: int) -> None:
        """Add a new system data signal to the DAQ configuration. The most
        common signal is SampleCollectionTime."""
        self.srcCode.set(code).wait()
        self._srcAdd.set("SYSTEM").wait()

    # Starting / stopping the DAQ
    _mode = Component(EpicsSignal, "MODE", kind=Kind.config, put_complete=True)
    _switch = Component(EpicsSignal, "SET", kind=Kind.omitted, put_complete=True)

    def start(self, mode=DataCollectionMode.Snapshot) -> None:
        """Start a new data collection"""
        self._mode.set(mode).wait()
        self._switch.set("START").wait()

    def stop(self) -> None:
        """Stop a running data collection"""
        self._switch.set("STOP").wait()

    def run(self, mode=DataCollectionMode.Snapshot) -> None:
        """Start a new data collection"""
        self._mode.set(mode).wait()
        self._switch.set("START").wait()
        # Wait for finishing acquisition
        # Note: this is very bad blocking sleep
        while self.status.value != 0:
            sleep(0.1)
        sleep(0.1)

        # Data readback
        data = self.data_rb.get()
        rows = self.data_rows.get()
        cols = self.data_cols.get()
        if len(data) == 0 or rows == 0 or cols == 0:
            sleep(0.5)
            data = self.data_rb.get()
            rows = self.data_rows.get()
            cols = self.data_cols.get()
        print(f"Data shape: {rows} x {cols}")
        data = data.reshape([int(rows), -1])
        return data

    # DAQ data readback
    data_rb = Component(EpicsPassiveRO, "DATA", kind=Kind.hinted)
    data_rows = Component(EpicsSignalRO, "DATA_ROWS", auto_monitor=True, kind=Kind.hinted)
    data_cols = Component(EpicsSignalRO, "DATA_COLS", auto_monitor=True, kind=Kind.hinted)
    data_stat = Component(EpicsSignalRO, "DATA_AVG", auto_monitor=True, kind=Kind.hinted)

    def dataReadBack(self) -> np.ndarray:
        """Retrieves collected data from the controller"""
        data = self.data_rb.get()
        rows = self.data_rows.get()
        cols = self.data_cols.get()
        if len(data) == 0 or rows == 0 or cols == 0:
            sleep(0.2)
            data = self.data_rb.get()
            rows = self.data_rows.get()
            cols = self.data_cols.get()
        print(f"Data shape: {rows} x {cols}")
        data = data.reshape([int(rows), -1])
        return data


class aa1GlobalVariables(Device):
    """Global variables

    This class provides an interface to directly read/write global variables
    on the Automation1 controller. These variables are accesible from script
    files and are thus a convenient way to interface with the outside word.

    Read operations take as input the memory address and the size
    Write operations work with the memory address and value

    Usage:
        var = aa1Tasks(AA1_IOC_NAME+":VAR:", name="var")
        var.wait_for_connection()
        ret = var.readInt(42)
        var.writeFloat(1000, np.random.random(1024))
        ret_arr = var.readFloat(1000, 1024)

    """

    # Status monitoring
    num_real = Component(EpicsSignalRO, "NUM-REAL_RBV", kind=Kind.config)
    num_int = Component(EpicsSignalRO, "NUM-INT_RBV", kind=Kind.config)
    num_string = Component(EpicsSignalRO, "NUM-STRING_RBV", kind=Kind.config)

    integer_addr = Component(EpicsSignal, "INT-ADDR", kind=Kind.omitted, put_complete=True)
    integer_size = Component(EpicsSignal, "INT-SIZE", kind=Kind.omitted, put_complete=True)
    integer = Component(EpicsSignal, "INT", kind=Kind.omitted, put_complete=True)
    integer_rb = Component(EpicsPassiveRO, "INT-RBV", kind=Kind.omitted)
    integerarr = Component(EpicsSignal, "INTARR", kind=Kind.omitted, put_complete=True)
    integerarr_rb = Component(EpicsPassiveRO, "INTARR-RBV", kind=Kind.omitted)

    real_addr = Component(EpicsSignal, "REAL-ADDR", kind=Kind.omitted, put_complete=True)
    real_size = Component(EpicsSignal, "REAL-SIZE", kind=Kind.omitted, put_complete=True)
    real = Component(EpicsSignal, "REAL", kind=Kind.omitted, put_complete=True)
    real_rb = Component(EpicsPassiveRO, "REAL-RBV", kind=Kind.omitted)
    realarr = Component(EpicsSignal, "REALARR", kind=Kind.omitted, put_complete=True)
    realarr_rb = Component(EpicsPassiveRO, "REALARR-RBV", kind=Kind.omitted)

    string_addr = Component(EpicsSignal, "STRING-ADDR", kind=Kind.omitted, put_complete=True)
    string = Component(EpicsSignal, "STRING", string=True, kind=Kind.omitted, put_complete=True)
    string_rb = Component(EpicsPassiveRO, "STRING-RBV", string=True, kind=Kind.omitted)

    def readInt(self, address: int, size: int = None) -> int:
        """Read a 64-bit integer global variable"""
        if address > self.num_int.get():
            raise RuntimeError("Integer address {address} is out of range")

        if size is None:
            self.integer_addr.set(address).wait()
            return self.integer_rb.get()
        else:
            self.integer_addr.set(address).wait()
            self.integer_size.set(size).wait()
            return self.integerarr_rb.get()

    def writeInt(self, address: int, value) -> None:
        """Write a 64-bit integer global variable"""
        if address > self.num_int.get():
            raise RuntimeError("Integer address {address} is out of range")

        if isinstance(value, (int, float)):
            self.integer_addr.set(address).wait()
            self.integer.set(value).wait()
        elif isinstance(value, np.ndarray):
            self.integer_addr.set(address).wait()
            self.integerarr.set(value).wait()
        elif isinstance(value, (list, tuple)):
            value = np.array(value, dtype=np.int32)
            self.integer_addr.set(address).wait()
            self.integerarr.set(value).wait()
        else:
            raise RuntimeError("Unsupported integer value type: {type(value)}")

    def readFloat(self, address: int, size: int = None) -> float:
        """Read a 64-bit double global variable"""
        if address > self.num_real.get():
            raise RuntimeError("Floating point address {address} is out of range")

        if size is None:
            self.real_addr.set(address).wait()
            return self.real_rb.get()
        else:
            self.real_addr.set(address).wait()
            self.real_size.set(size).wait()
            return self.realarr_rb.get()

    def writeFloat(self, address: int, value) -> None:
        """Write a 64-bit float global variable"""
        if address > self.num_real.get():
            raise RuntimeError("Float address {address} is out of range")

        if isinstance(value, (int, float)):
            self.real_addr.set(address).wait()
            self.real.set(value).wait()
        elif isinstance(value, np.ndarray):
            self.real_addr.set(address).wait()
            self.realarr.set(value).wait()
        elif isinstance(value, (list, tuple)):
            value = np.array(value)
            self.real_addr.set(address).wait()
            self.realarr.set(value).wait()
        else:
            raise RuntimeError("Unsupported float value type: {type(value)}")

    def readString(self, address: int) -> str:
        """Read a 40 letter string global variable
        ToDo: Automation 1 strings are 256 bytes
        """
        if address > self.num_string.get():
            raise RuntimeError("String address {address} is out of range")

        self.string_addr.set(address).wait()
        return self.string_rb.get()

    def writeString(self, address: int, value) -> None:
        """Write a 40 bytes string  global variable"""
        if address > self.num_string.get():
            raise RuntimeError("Integer address {address} is out of range")

        if isinstance(value, str):
            self.string_addr.set(address).wait()
            self.string.set(value).wait()
        else:
            raise RuntimeError("Unsupported string value type: {type(value)}")


class aa1GlobalVariableBindings(Device):
    """Polled global variables

    This class provides an interface to read/write the first few global variables
    on the Automation1 controller. These variables are continuously polled
    and are thus a convenient way to interface scripts with the outside word.
    """

    int0 = Component(EpicsSignalRO, "INT0_RBV", auto_monitor=True, name="int0", kind=Kind.hinted)
    int1 = Component(EpicsSignalRO, "INT1_RBV", auto_monitor=True, name="int1", kind=Kind.hinted)
    int2 = Component(EpicsSignalRO, "INT2_RBV", auto_monitor=True, name="int2", kind=Kind.hinted)
    int3 = Component(EpicsSignalRO, "INT3_RBV", auto_monitor=True, name="int3", kind=Kind.hinted)
    int8 = Component(
        EpicsSignal,
        "INT8_RBV",
        put_complete=True,
        write_pv="INT8",
        auto_monitor=True,
        name="int8",
        kind=Kind.hinted,
    )
    int9 = Component(
        EpicsSignal,
        "INT9_RBV",
        put_complete=True,
        write_pv="INT9",
        auto_monitor=True,
        name="int9",
        kind=Kind.hinted,
    )
    int10 = Component(
        EpicsSignal,
        "INT10_RBV",
        put_complete=True,
        write_pv="INT10",
        auto_monitor=True,
        name="int10",
        kind=Kind.hinted,
    )
    int11 = Component(
        EpicsSignal,
        "INT11_RBV",
        put_complete=True,
        write_pv="INT11",
        auto_monitor=True,
        name="int11",
        kind=Kind.hinted,
    )

    float0 = Component(
        EpicsSignalRO, "REAL0_RBV", auto_monitor=True, name="float0", kind=Kind.hinted
    )
    float1 = Component(
        EpicsSignalRO, "REAL1_RBV", auto_monitor=True, name="float1", kind=Kind.hinted
    )
    float2 = Component(
        EpicsSignalRO, "REAL2_RBV", auto_monitor=True, name="float2", kind=Kind.hinted
    )
    float3 = Component(
        EpicsSignalRO, "REAL3_RBV", auto_monitor=True, name="float3", kind=Kind.hinted
    )
    float16 = Component(
        EpicsSignal,
        "REAL16_RBV",
        write_pv="REAL16",
        put_complete=True,
        auto_monitor=True,
        name="float16",
        kind=Kind.hinted,
    )
    float17 = Component(
        EpicsSignal,
        "REAL17_RBV",
        write_pv="REAL17",
        put_complete=True,
        auto_monitor=True,
        name="float17",
        kind=Kind.hinted,
    )
    float18 = Component(
        EpicsSignal,
        "REAL18_RBV",
        write_pv="REAL18",
        put_complete=True,
        auto_monitor=True,
        name="float18",
        kind=Kind.hinted,
    )
    float19 = Component(
        EpicsSignal,
        "REAL19_RBV",
        write_pv="REAL19",
        put_complete=True,
        auto_monitor=True,
        name="float19",
        kind=Kind.hinted,
    )

    # BEC LiveTable crashes on non-numeric values
    str0 = Component(
        EpicsSignalRO, "STR0_RBV", auto_monitor=True, string=True, name="str0", kind=Kind.config
    )
    str1 = Component(
        EpicsSignalRO, "STR1_RBV", auto_monitor=True, string=True, name="str1", kind=Kind.config
    )
    str4 = Component(
        EpicsSignal,
        "STR4_RBV",
        put_complete=True,
        string=True,
        auto_monitor=True,
        write_pv="STR4",
        name="str4",
        kind=Kind.config,
    )
    str5 = Component(
        EpicsSignal,
        "STR5_RBV",
        put_complete=True,
        string=True,
        auto_monitor=True,
        write_pv="STR5",
        name="str5",
        kind=Kind.config,
    )


class aa1AxisIo(Device):
    """Analog / digital Input-Output

    This class provides convenience wrappers around the Aerotech API's axis
    specific IO functionality. Note that this is a low-speed API, actual work
    should be done in AeroScript. Only one pin can be writen directly but
    several can be polled!
    """

    polllvl = Component(EpicsSignal, "POLLLVL", put_complete=True, kind=Kind.config)
    ai0 = Component(EpicsSignalRO, "AI0-RBV", auto_monitor=True, kind=Kind.hinted)
    ai1 = Component(EpicsSignalRO, "AI1-RBV", auto_monitor=True, kind=Kind.hinted)
    ai2 = Component(EpicsSignalRO, "AI2-RBV", auto_monitor=True, kind=Kind.hinted)
    ai3 = Component(EpicsSignalRO, "AI3-RBV", auto_monitor=True, kind=Kind.hinted)
    ao0 = Component(EpicsSignalRO, "AO0-RBV", auto_monitor=True, kind=Kind.hinted)
    ao1 = Component(EpicsSignalRO, "AO1-RBV", auto_monitor=True, kind=Kind.hinted)
    ao2 = Component(EpicsSignalRO, "AO2-RBV", auto_monitor=True, kind=Kind.hinted)
    ao3 = Component(EpicsSignalRO, "AO3-RBV", auto_monitor=True, kind=Kind.hinted)
    di0 = Component(EpicsSignalRO, "DI0-RBV", auto_monitor=True, kind=Kind.hinted)
    do0 = Component(EpicsSignalRO, "DO0-RBV", auto_monitor=True, kind=Kind.hinted)

    ai_addr = Component(EpicsSignal, "AI-ADDR", put_complete=True, kind=Kind.config)
    ai = Component(EpicsSignalRO, "AI-RBV", auto_monitor=True, kind=Kind.hinted)

    ao_addr = Component(EpicsSignal, "AO-ADDR", put_complete=True, kind=Kind.config)
    ao = Component(EpicsSignal, "AO-RBV", write_pv="AO", auto_monitor=True, kind=Kind.hinted)

    di_addr = Component(EpicsSignal, "DI-ADDR", put_complete=True, kind=Kind.config)
    di = Component(EpicsSignalRO, "DI-RBV", auto_monitor=True, kind=Kind.hinted)

    do_addr = Component(EpicsSignal, "DO-ADDR", put_complete=True, kind=Kind.config)
    do = Component(EpicsSignal, "DO-RBV", write_pv="DO", auto_monitor=True, kind=Kind.hinted)

    def setAnalog(self, pin: int, value: float, settle_time=0.05):
        # Set the address
        self.ao_addr.set(pin).wait()
        # Set the voltage
        self.ao.set(value, settle_time=settle_time).wait()

    def setDigital(self, pin: int, value: int, settle_time=0.05):
        # Set the address
        self.do_addr.set(pin).wait()
        # Set the voltage
        self.do.set(value, settle_time=settle_time).wait()


class aa1AxisPsoBase(Device):
    """Position Sensitive Output - Base class

    This class provides convenience wrappers around the Aerotech IOC's PSO
    functionality. As a base class, it's just a collection of PVs without
    significant logic (that should be implemented in the child classes).
    It uses event-waveform concept to produce signals on the configured
    output pin: a specified position based event will trigger the generation
    of a waveform on the oputput that can be either used as exposure enable,
    as individual trigger or as a series of triggers per each event.
    As a first approach, the module follows a simple pipeline structure:
        Genrator --> Event --> Waveform --> Output

    Specific operation modes should be implemented in child classes.
    """

    # ########################################################################
    # General module status
    status = Component(EpicsSignalRO, "STATUS", auto_monitor=True, kind=Kind.hinted)
    output = Component(EpicsSignalRO, "OUTPUT-RBV", auto_monitor=True, kind=Kind.hinted)
    _eventSingle = Component(EpicsSignal, "EVENT:SINGLE", put_complete=True, kind=Kind.omitted)
    _reset = Component(EpicsSignal, "RESET", put_complete=True, kind=Kind.omitted)
    posInput = Component(EpicsSignal, "DIST:INPUT", put_complete=True, kind=Kind.omitted)

    # ########################################################################
    # PSO Distance event module
    dstEventsEna = Component(EpicsSignal, "DIST:EVENTS", put_complete=True, kind=Kind.omitted)
    dstCounterEna = Component(EpicsSignal, "DIST:COUNTER", put_complete=True, kind=Kind.omitted)
    dstCounterVal = Component(EpicsSignalRO, "DIST:CTR0_RBV", auto_monitor=True, kind=Kind.hinted)
    dstArrayIdx = Component(EpicsSignalRO, "DIST:IDX_RBV", auto_monitor=True, kind=Kind.hinted)
    dstArrayDepleted = Component(
        EpicsSignalRO, "DIST:ARRAY-DEPLETED-RBV", auto_monitor=True, kind=Kind.hinted
    )

    dstDirection = Component(EpicsSignal, "DIST:EVENTDIR", put_complete=True, kind=Kind.omitted)
    dstDistance = Component(EpicsSignal, "DIST:DISTANCE", put_complete=True, kind=Kind.hinted)
    dstDistanceArr = Component(EpicsSignal, "DIST:DISTANCES", put_complete=True, kind=Kind.omitted)
    dstArrayRearm = Component(EpicsSignal, "DIST:REARM-ARRAY", put_complete=True, kind=Kind.omitted)

    # ########################################################################
    # PSO Window event module
    winEvents = Component(EpicsSignal, "WINDOW:EVENTS", put_complete=True, kind=Kind.omitted)
    winOutput = Component(EpicsSignal, "WINDOW0:OUTPUT", put_complete=True, kind=Kind.omitted)
    winInput = Component(EpicsSignal, "WINDOW0:INPUT", put_complete=True, kind=Kind.omitted)
    winCounter = Component(EpicsSignal, "WINDOW0:COUNTER", put_complete=True, kind=Kind.omitted)
    _winLower = Component(EpicsSignal, "WINDOW0:LOWER", put_complete=True, kind=Kind.omitted)
    _winUpper = Component(EpicsSignal, "WINDOW0:UPPER", put_complete=True, kind=Kind.omitted)

    # ########################################################################
    # PSO waveform module
    waveEnable = Component(EpicsSignal, "WAVE:ENABLE", put_complete=True, kind=Kind.omitted)
    waveMode = Component(EpicsSignal, "WAVE:MODE", put_complete=True, kind=Kind.omitted)
    # waveDelay = Component(EpicsSignal, "WAVE:DELAY", put_complete=True, kind=Kind.omitted)

    # PSO waveform pulse output
    # pulseTrunc = Component(EpicsSignal, "WAVE:PULSE:TRUNC", put_complete=True, kind=Kind.omitted)
    pulseOnTime = Component(EpicsSignal, "WAVE:PULSE:ONTIME", put_complete=True, kind=Kind.omitted)
    pulseWindow = Component(EpicsSignal, "WAVE:PULSE:PERIOD", put_complete=True, kind=Kind.omitted)
    pulseCount = Component(EpicsSignal, "WAVE:PULSE:COUNT", put_complete=True, kind=Kind.omitted)
    pulseApply = Component(EpicsSignal, "WAVE:PULSE:APPLY", put_complete=True, kind=Kind.omitted)

    # ########################################################################
    # PSO output module
    outPin = Component(EpicsSignal, "PIN", put_complete=True, kind=Kind.omitted)
    outSource = Component(EpicsSignal, "SOURCE", put_complete=True, kind=Kind.omitted)

    def fire(self, settle_time=None):
        """Fire a single PSO event (i.e. manual software trigger)"""
        self._eventSingle.set(1, settle_time=settle_time).wait()

    def toggle(self):
        orig_waveMode = self.waveMode.get()
        self.waveMode.set("Toggle").wait()
        self.fire(0.1)
        self.waveMode.set(orig_waveMode).wait()


class aa1AxisPsoDistance(aa1AxisPsoBase):
    """Position Sensitive Output - Distance mode

    This class provides convenience wrappers around the Aerotech API's PSO
    functionality in distance mode. It uses event-waveform concept to produce
    signals on the configured output pin: a specified position based event
    will trigger the generation af a waveform on the oputput that can be either
    used as exposure enable, as individual trigger or as a series of triggers
    per each event.
    As a first approach, the module follows a simple pipeline structure:
        Genrator (distance) --> Event --> Waveform --> Output

    The module provides configuration interface to common functionality, such
    as fixed distance or array based triggering and can serve as a base for
    future advanced functionality. The relative distances ease the limitations
    coming from 32 bit PSO positions.
    For a more detailed description of additional signals and masking plase
    refer to Automation1's online manual.

    Usage:
        # Configure the PSO to raise an 'enable' signal for 180 degrees
        pso = aa1AxisPsoDistance(AA1_IOC_NAME+":ROTY:PSO:", name="pso")
        pso.wait_for_connection()
        pso.configure(d={'distance': [180, 0.1], 'wmode': "toggle"})
        pso.kickoff().wait()

        # Configure the PSO to emmit 5 triggers every 1.8 degrees
        pso = aa1AxisPsoDistance(AA1_IOC_NAME+":ROTY:PSO:", name="pso")
        pso.wait_for_connection()
        pso.configure(d={'distance': 1.8, 'wmode': "pulsed", 'n_pulse': 5})
        pso.kickoff().wait()
    """

    SUB_PROGRESS = "progress"

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        """__init__ MUST have a full argument list"""
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self._Vdistance = 3.141592
        self.subscribe(self._progress_update, "progress", run=False)

    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan"""
        if self.dstArrayDepleted.value:
            print("PSO array depleted")
            self._run_subs(sub_type=self.SUB_PROGRESS, value=1, max_value=1, done=1)
            return

        progress = 1
        max_value = 1
        print(f"PSO array proggress: {progress}")
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=int(progress),
            max_value=max_value,
            done=int(np.isclose(max_value, progress, 1e-3)),
        )

    # ########################################################################
    # PSO high level interface
    def configure(self, d: dict = {}) -> tuple:
        """Simplified configuration interface to access the most common
        functionality for distance mode PSO.

        :param distance: The trigger distance or the array of distances between subsequent points.
        :param wmode: Waveform mode configuration, usually pulsed/toggled.
        """
        distance = d["distance"]
        wmode = str(d["wmode"])
        t_pulse = float(d["t_pulse"]) if "t_pulse" in d else 100
        w_pulse = float(d["w_pulse"]) if "w_pulse" in d else 200
        n_pulse = float(d["n_pulse"]) if "n_pulse" in d else 1

        # Validate input parameters
        if wmode not in ["pulse", "pulsed", "toggle", "toggled"]:
            raise RuntimeError(f"Unsupported distace triggering mode: {wmode}")

        old = self.read_configuration()
        # Configure distance generator (also resets counter to 0)
        self._distanceValue = distance
        if isinstance(distance, (float, int)):
            self.dstDistance.set(distance).wait()
        elif isinstance(distance, (np.ndarray, list, tuple)):
            self.dstDistanceArr.set(distance).wait()

        self.winEvents.set("Off").wait()
        self.dstCounterEna.set("Off").wait()
        self.dstEventsEna.set("Off").wait()

        # Configure the pulsed/toggled waveform
        if wmode in ["toggle", "toggled"]:
            # Switching to simple toggle mode
            self.waveEnable.set("On").wait()
            self.waveMode.set("Toggle").wait()

        elif wmode in ["pulse", "pulsed"]:
            # Switching to pulsed mode
            self.waveEnable.set("On").wait()
            self.waveMode.set("Pulse").wait()
            # Setting pulse shape
            if w_pulse is not None:
                self.pulseWindow.set(w_pulse).wait()
            if t_pulse is not None:
                self.pulseOnTime.set(t_pulse).wait()
            if n_pulse is not None:
                self.pulseCount.set(n_pulse).wait()
            # Commiting configuration
            self.pulseApply.set(1).wait()
            # Enabling PSO waveform outputs
            self.waveEnable.set("On").wait()
        else:
            raise RuntimeError(f"Unsupported waveform mode: {wmode}")

        # Ensure output is set to low
        if self.output.value:
            self.toggle()

        # Set PSO output data source
        self.outSource.set("Waveform").wait()

        new = self.read_configuration()
        print("PSO configured")
        return (old, new)

    # ########################################################################
    # Bluesky step scan interface
    def complete(self, settle_time=0.1) -> DeviceStatus:
        """DDC just reads back whatever is available in the buffers"""
        sleep(settle_time)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def trigger(self):
        return super().trigger()

    def unstage(self):
        # Turn off counter monitoring
        self.dstEventsEna.set("Off").wait()
        self.dstCounterEna.set("Off").wait()
        return super().unstage()

    # ########################################################################
    # Bluesky flyer interface
    def kickoff(self) -> DeviceStatus:
        # Rearm the configured array
        if hasattr(self, "_distanceValue") and isinstance(
            self._distanceValue, (np.ndarray, list, tuple)
        ):
            self.dstArrayRearm.set(1).wait()
        # Start monitoring the counters
        self.dstEventsEna.set("On").wait()
        self.dstCounterEna.set("On").wait()
        status = DeviceStatus(self)
        status.set_finished()
        print("PSO kicked off")
        return status

    # def complete(self) -> DeviceStatus:
    #     """Bluesky flyer interface"""
    #     # Array mode waits until the buffer is empty
    #     if hasattr(self, "_distanceValue") and isinstance(
    #         self._distanceValue, (np.ndarray, list, tuple)
    #     ):
    #         # Define wait until the busy flag goes down (excluding initial update)
    #         timestamp_ = 0

    #         def notRunning(*args, old_value, value, timestamp, **kwargs):
    #             nonlocal timestamp_
    #             result = False if (timestamp_ == 0) else bool(int(value) & 0x1000)
    #             print(f"Old {old_value}\tNew: {value}\tResult: {result}")
    #             timestamp_ = timestamp
    #             return result

    #         # Subscribe and wait for update
    #         # status = SubscriptionStatus(self.status, notRunning, settle_time=0.5)
    #         # Data capture can be stopped any time
    #         status = DeviceStatus(self)
    #         status.set_finished()
    #     else:
    #         # In distance trigger mode there's no specific goal
    #         status = DeviceStatus(self)
    #         status.set_finished()
    #     return status

    def describe_collect(self) -> OrderedDict:
        ret = OrderedDict()
        ret["index"] = {
            "source": "internal",
            "dtype": "integer",
            "shape": [],
            "units": "",
            "lower_ctrl_limit": 0,
            "upper_ctrl_limit": 0,
        }
        return {self.name: ret}

    def collect(self) -> OrderedDict:
        ret = OrderedDict()
        ret["timestamps"] = {"index": time.time()}
        ret["data"] = {"index": self.dstCounterVal.value}
        yield ret


class aa1AxisPsoWindow(aa1AxisPsoBase):
    """Position Sensitive Output - Window mode

    This class provides convenience wrappers around the Aerotech API's PSO
    functionality in window mode. It can either use the event-waveform concept
    or provide a direct window output signal (in/out) to the output pin. The
    latter is particularly well-suited for the generation of trigger enable
    signals, while in event mode it allows the finetuning of trigger lenth.
    As a first approach, the module follows a simple pipeline structure:
        Genrator --> Event --> Waveform --> Output pin
        Genrator --> Window output --> Output pin

    The module provides configuration interface to common functionality, such
    as repeated trigger enable signal or fixed area scaning. Unfortunately the
    entered positions are absolute, meaning this mode has an inherent limitation
    with encoder counters being kept in 32 bit integers.
    For a more detailed description of additional signals and masking plase
    refer to Automation1's online manual.
    """

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        """__init__ MUST have a full argument list"""
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self._mode = "output"
        self._eventMode = "Enter"

    # ########################################################################
    # PSO high level interface
    def configure(self, d: dict = {}) -> tuple:
        """Simplified configuration interface to access the most common
        functionality for distance mode PSO.

        :param distance: The trigger distance or the array of distances between subsequent points.
        :param wmode: Waveform mode configuration, usually output/pulsed/toggled.
        """
        bounds = d["distance"]
        wmode = str(d["wmode"])
        emode = str(d["emode"])
        t_pulse = float(d["t_pulse"]) if "t_pulse" in d else None
        w_pulse = float(d["w_pulse"]) if "w_pulse" in d else None
        n_pulse = float(d["n_pulse"]) if "n_pulse" in d else None

        # Validate input parameters
        if wmode not in ["pulse", "pulsed", "toggle", "toggled", "output", "flag"]:
            raise RuntimeError(f"Unsupported window triggering mode: {wmode}")
        self._mode = wmode
        self._eventMode = emode

        old = self.read_configuration()

        # Configure the window module
        # Set the window ranges (MUST be in start position)
        if len(bounds) == 2:
            self.winCounter.set(0).wait()
            self._winLower.set(bounds[0]).wait()
            self._winUpper.set(bounds[1]).wait()

        elif isinstance(bounds, np.ndarray):
            # ToDo...
            pass

        # Don't start triggering just yet
        self.winOutput.set("Off").wait()
        self.winEvents.set("Off").wait()

        # Configure the pulsed/toggled waveform
        if wmode in ["toggle", "toggled"]:
            # Switching to simple toggle mode
            self.waveEnable.set("On").wait()
            self.waveMode.set("Toggle").wait()
        elif wmode in ["pulse", "pulsed"]:
            # Switching to pulsed mode
            self.waveEnable.set("On").wait()
            self.waveMode.set("Pulse").wait()
            # Setting pulse shape
            self.pulseWindow.set(w_pulse).wait()
            self.pulseOnTime.set(t_pulse).wait()
            self.pulseCount.set(n_pulse).wait()
            # Commiting configuration
            self.pulseApply.set(1).wait()
            # Enabling PSO waveform outputs
            self.waveEnable.set("On").wait()
        elif wmode in ["output", "flag"]:
            self.waveEnable.set("Off").wait()
        else:
            raise RuntimeError(f"Unsupported window mode: {wmode}")

        # Set PSO output data source
        if wmode in ["toggle", "toggled", "pulse", "pulsed"]:
            self.outSource.set("Waveform").wait()
        elif wmode in ["output", "flag"]:
            self.outSource.set("Window").wait()

        new = self.read_configuration()
        return (old, new)

    def stage(self, settle_time=None):
        if self.outSource.get() in ["Window", 2]:
            self.winOutput.set("On").wait()
        else:
            self.winEvents.set(self._eventMode).wait()
        if settle_time is not None:
            sleep(settle_time)
        return super().stage()

    def kickoff(self, settle_time=None):
        if self.outSource.get() in ["Window", 2]:
            self.winOutput.set("On").wait()
        else:
            self.winEvents.set(self._eventMode).wait()
        if settle_time is not None:
            sleep(settle_time)

    def unstage(self, settle_time=None):
        self.winOutput.set("Off").wait()
        self.winEvents.set("Off").wait()
        if settle_time is not None:
            sleep(settle_time)
        return super().unstage()


class aa1AxisDriveDataCollection(Device):
    """Axis data collection

    This class provides convenience wrappers around the Aerotech API's axis
    specific data collection functionality. This module allows to record
    hardware synchronized signals with up to 200 kHz.

    The default configuration is using a fixed memory mapping allowing up to
    1 million recorded data points on an XC4e (this depends on controller).

    Usage:
        # Configure the DDC with default internal triggers
        ddc = aa1AxisPsoDistance(AA1_IOC_NAME+":ROTY:DDC:", name="ddc")
        ddc.wait_for_connection()
        ddc.configure(d={'npoints': 5000})
        ddc.kickoff().wait()
        ...
        ret = yield from ddc.collect()
    """

    # ########################################################################
    # General module status
    state = Component(EpicsSignalRO, "STATE", auto_monitor=True, kind=Kind.hinted)
    nsamples_rbv = Component(EpicsSignalRO, "SAMPLES_RBV", auto_monitor=True, kind=Kind.hinted)
    _switch = Component(EpicsSignal, "ACQUIRE", put_complete=True, kind=Kind.omitted)
    _input0 = Component(EpicsSignal, "INPUT0", put_complete=True, kind=Kind.config)
    _input1 = Component(EpicsSignal, "INPUT1", put_complete=True, kind=Kind.config)
    _trigger = Component(EpicsSignal, "TRIGGER", put_complete=True, kind=Kind.config)

    npoints = Component(EpicsSignal, "NPOINTS", put_complete=True, kind=Kind.config)
    _readback0 = Component(EpicsSignal, "AREAD0", kind=Kind.omitted)
    _readstatus0 = Component(EpicsSignalRO, "AREAD0_RBV", auto_monitor=True, kind=Kind.omitted)
    _readback1 = Component(EpicsSignal, "AREAD1", kind=Kind.omitted)
    _readstatus1 = Component(EpicsSignalRO, "AREAD1_RBV", auto_monitor=True, kind=Kind.omitted)

    _buffer0 = Component(EpicsSignalRO, "BUFFER0", auto_monitor=True, kind=Kind.hinted)
    _buffer1 = Component(EpicsSignalRO, "BUFFER1", auto_monitor=True, kind=Kind.hinted)

    SUB_PROGRESS = "progress"

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.subscribe(self._progress_update, "progress", run=False)

    def _progress_update(self, value, **kwargs) -> None:
        """Progress update on the scan"""
        if self.state.value not in (2, "Acquiring"):
            self._run_subs(sub_type=self.SUB_PROGRESS, value=1, max_value=1, done=1)
            return

        progress = 1
        max_value = 1
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=int(progress),
            max_value=max_value,
            done=int(np.isclose(max_value, progress, 1e-3)),
        )

    def configure(self, d: dict = {}) -> tuple:
        npoints = int(d["npoints"])
        trigger = int(d["trigger"]) if "trigger" in d else DriveDataCaptureTrigger.PsoOutput
        source0 = int(d["source0"]) if "source0" in d else DriveDataCaptureInput.PrimaryFeedback
        source1 = int(d["source1"]) if "source1" in d else DriveDataCaptureInput.PositionCommand

        old = self.read_configuration()

        self._input0.set(source0).wait()
        self._input1.set(source1).wait()
        self._trigger.set(trigger).wait()
        # This allocates the memory...
        self.npoints.set(npoints).wait()

        new = self.read_configuration()
        return (old, new)

    # Bluesky step scanning interface
    def stage(self, settle_time=0.1):
        self._switch.set("Start", settle_time=0.5).wait()
        return super().stage()

    def unstage(self, settle_time=0.1):
        self._switch.set("Stop", settle_time=settle_time).wait()
        super().unstage()
        print(f"Recorded samples: {self.nsamples_rbv.value}")

    # Bluesky flyer interface
    def kickoff(self, settle_time=0.1) -> Status:
        status = self._switch.set("Start", settle_time=settle_time)
        return status

    def complete(self, settle_time=0.1) -> DeviceStatus:
        """DDC just reads back whatever is available in the buffers"""
        sleep(settle_time)
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def _collect(self, index=0):
        """Force a readback of the data buffer

        Note that there's a weird behaviour in ophyd that it issues an
        initial update event with the initial value but 0 timestamp. Theese
        old_values are invalid and must be filtered out.
        """

        # Define wait until the busy flag goes down (excluding initial update)
        timestamp_ = 0

        def negEdge(*args, old_value, value, timestamp, **kwargs):
            nonlocal timestamp_
            result = False if (timestamp_ == 0) else (old_value == 1 and value == 0)
            # print(f"\nBuffer1 status:\t{old_value} ({timestamp_}) to  {value} ({timestamp}) Result: {result}")
            timestamp_ = timestamp
            return result

        if index == 0:
            status = SubscriptionStatus(self._readstatus0, negEdge, settle_time=0.5)
            self._readback0.set(1).wait()
        elif index == 1:
            status = SubscriptionStatus(self._readstatus1, negEdge, settle_time=0.5)
            self._readback1.set(1).wait()

        # Start asynchronous readback
        status.wait()
        return status

    def describe_collect(self) -> OrderedDict:
        ret = OrderedDict()
        ret["buffer0"] = {
            "source": "internal",
            "dtype": "array",
            "shape": [],
            "units": "",
            "lower_ctrl_limit": 0,
            "upper_ctrl_limit": 0,
        }
        ret["buffer1"] = {
            "source": "internal",
            "dtype": "array",
            "shape": [],
            "units": "",
            "lower_ctrl_limit": 0,
            "upper_ctrl_limit": 0,
        }
        return {self.name: ret}

    def collect(self) -> OrderedDict:

        self._collect(0).wait()
        self._collect(1).wait()

        b0 = self._buffer0.value
        b1 = self._buffer1.value
        ret = OrderedDict()
        ret["timestamps"] = {"buffer0": time.time(), "buffer1": time.time()}
        ret["data"] = {"buffer0": b0, "buffer1": b1}
        yield ret


# Automatically start simulation if directly invoked
if __name__ == "__main__":
    AA1_IOC_NAME = "X02DA-ES1-SMP1"
    AA1_AXIS_NAME = "ROTY"
    # Drive data collection
    task = aa1Tasks(AA1_IOC_NAME + ":TASK:", name="tsk")
    task.wait_for_connection()
    task.describe()
    ddc = aa1AxisDriveDataCollection("X02DA-ES1-SMP1:ROTY:DDC:", name="ddc")
    ddc.wait_for_connection()
    globb = aa1GlobalVariableBindings(AA1_IOC_NAME + ":VAR:", name="globb")
    globb.wait_for_connection()
    globb.describe()
    mot = EpicsMotor(AA1_IOC_NAME + ":" + AA1_AXIS_NAME, name="x")
    mot.wait_for_connection()



from unittest import mock

import pytest
from utils import SocketMock

from ophyd_devices.galil.fupr_ophyd import FuprGalilController, FuprGalilMotor


@pytest.fixture
def fsamroy():
    FuprGalilController._reset_controller()
    fsamroy_motor = FuprGalilMotor(
        "A", name="fsamroy", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
    )
    fsamroy_motor.controller.on()
    assert isinstance(fsamroy_motor.controller, FuprGalilController)
    yield fsamroy_motor
    fsamroy_motor.controller.off()
    fsamroy_motor.controller._reset_controller()


@pytest.mark.parametrize(
    "pos,msg_received,msg_put,sign",
    [
        (-0.5, b" -12800\n\r", [b"TPA\r", b"MG_BGA\r", b"TPA\r"], 1),
        (-0.5, b" 12800\n\r", [b"TPA\r", b"MG_BGA\r", b"TPA\r"], -1),
    ],
)
def test_axis_get(fsamroy, pos, msg_received, msg_put, sign):
    fsamroy.sign = sign
    fsamroy.device_manager = mock.MagicMock()
    fsamroy.controller.sock.flush_buffer()
    fsamroy.controller.sock.buffer_recv = msg_received
    val = fsamroy.read()
    assert val["fsamroy"]["value"] == pos
    assert fsamroy.readback.get() == pos
    assert fsamroy.controller.sock.buffer_put == msg_put


@pytest.mark.parametrize(
    "target_pos,socket_put_messages,socket_get_messages",
    [
        (
            0,
            [b"MG axisref\r", b"PAA=0\r", b"PAA=0\r", b"BGA\r"],
            [b"1.00", b"-1", b":", b":", b":", b":", b"-1"],
        )
    ],
)
def test_axis_put(fsamroy, target_pos, socket_put_messages, socket_get_messages):
    fsamroy.controller.sock.flush_buffer()
    fsamroy.controller.sock.buffer_recv = socket_get_messages
    fsamroy.user_setpoint.put(target_pos)
    assert fsamroy.controller.sock.buffer_put == socket_put_messages


def test_drive_axis_to_limit(fsamroy):
    fsamroy.controller.sock.flush_buffer()
    with pytest.raises(NotImplementedError):
        fsamroy.controller.drive_axis_to_limit(0, "forward")



""" This module provides a range of protocols that describe the expected interface for different types of devices.

The protocols below can be used as teamplates for functionality to be implemeted by different type of devices.
They further facilitate runtime checks on devices and provide a minimum set of properties required for a device to be loadable by BEC.

The protocols are:
- BECDeviceProtocol: Protocol for devices in BEC. All devices must at least implement this protocol.
- BECSignalProtocol: Protocol for signals.
- BECScanProtocol: Protocol for the scan interface.
- BECMixinProtocol: Protocol for utilities in particular relevant for detector implementations.
- BECPositionerProtocol: Protocol for positioners.
- BECFlyerProtocol: Protocol with for flyers.

Keep in mind, that a device of type flyer should generally also implement the BECScanProtocol that provides the required functionality for scans.
Flyers in addition, also implement the BECFlyerProtocol. Similarly, positioners should also implement the BECScanProtocol and BECPositionerProtocol.

"""

from typing import Protocol, runtime_checkable

from bec_lib.file_utils import FileWriter
from ophyd import Component, DeviceStatus, Kind, Staged

from ophyd_devices.utils import bec_scaninfo_mixin


@runtime_checkable
class BECDeviceProtocol(Protocol):
    """Protocol for ophyd objects with zero functionality."""

    _destroyed: bool

    @property
    def name(self) -> str:
        """name property"""

    @name.setter
    def name(self, value: str) -> None:
        """name setter"""

    @property
    def kind(self) -> Kind:
        """kind property"""

    @kind.setter
    def kind(self, value: Kind):
        """kind setter"""

    @property
    def parent(self) -> object:
        """Property to find the parent device"""

    @property
    def root(self) -> object:
        """Property to fint the root device"""

    @property
    def hints(self) -> dict:
        """hints property"""

    @property
    def connected(self) -> bool:
        """connected property.
        Check if signals are connected

        Returns:
            bool: True if connected, False otherwise
        """

    @connected.setter
    def connected(self, value: bool):
        """connected setter"""

    def read(self) -> dict:
        """read method

        Override by child class with read method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.normal or kind.hinted:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """

    def read_configuration(self) -> dict:
        """read_configuration method

        Override by child class with read_configuration method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.config:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """

    def describe(self) -> dict:
        """describe method

        Override by child class with describe method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """

    def describe_configuration(self) -> dict:
        """describe method

        Includes all signals of type Kind.config.
        Override by child class with describe_configuration method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """

    def destroy(self) -> None:
        """Destroy method.

        _destroyed must be set to True after calling destroy.
        """

    def trigger(self) -> DeviceStatus:
        """Trigger method on the device

        Returns ophyd DeviceStatus object, which is used to track the status of the trigger.
        It can also be blocking until the trigger is completed, and return the status object
        with set_finished() method called on the DeviceStatus.
        """


@runtime_checkable
class BECSignalProtocol(Protocol):
    """Protocol for BEC signals with zero functionality.

    This protocol adds the specific implementation for a signal.
    Please be aware that a signal must also implement BECDeviceProtocol.

    Note: Currently the implementation of the protocol is not taking into account the
    event_model from ophyd, i.e. _run_sbus
    """

    @property
    def limits(self) -> tuple[float, float]:
        """Limits property for signals.
        If low_limit == high_limit, it is equivalent to NO limits!

        Returns:
            tuple: Tuple with lower and upper limits
        """

    @property
    def high_limit(self) -> float:
        """High limit property for signals.

        Returns:
            float: Upper limit
        """

    @property
    def low_limit(self) -> float:
        """Low limit property for signals.

        Returns:
            float: Lower limit
        """

    @property
    def write_access(self) -> bool:
        """Write access method for signals.

        Returns:
            bool: True if write access is allowed, False otherwise
        """

    def check_value(self, value: float) -> None:
        """Check whether value is within limits

        Args:
            value: value to check

        Raises:
            LimitError in case the requested motion is not inside of limits.
        """

    def put(self, value: any, force: bool = False, timeout: float = None):
        """Put method for signals.
        This method should resolve immediately and not block.
        If not force, the method checks if the value is within limits using check_value.


        Args:
            value (any)     : value to put
            force (bool)    : Flag to force the put and ignore limits
            timeout (float) : Timeout for the put
        """

    def set(self, value: any, timeout: float = None) -> DeviceStatus:
        """Set method for signals.
        This method should be blocking until the set is completed.

        Args:
            value (any)     : value to set
            timeout (float) : Timeout for the set

        Returns:
            DeviceStatus    : DeviceStatus object that will finish upon return
        """


@runtime_checkable
class BECScanProtocol(BECDeviceProtocol, Protocol):
    """Protocol for devices offering an Protocol with all relevant functionality for scans.

    In BEC, scans typically follow the order of stage, (pre_scan), trigger, unstage.
    Stop should be used to interrupt a scan. Be aware that pre_scan is optional and therefor
    part of the BECMixinProtocol, typically useful for more complex devices such as detectors.

    This protocol allows to perform runtime checks on devices of ophyd.
    It is the minimum set of properties required for a device to be loadable by BEC.
    """

    _staged: Staged
    """Staged property to indicate if the device is staged."""

    def stage(self) -> list[object]:
        """Stage method to prepare the device for an upcoming acquistion.

        This prepares a device for an upcoming acquisition, i.e. it is the first
        method for which the scan parameters are known and the device can be configured.

        It can be used to move scan_motors to their start position
        or also prepare DAQ systems for the upcoming measurement.
        We can further publish the file location for DAQ systems
        to BEC and inform BEC's file writer where data will be written to.

        Stagin is not idempotent. If called twice without an unstage it should raise.
        For ophyd devices, one may used self._staged = True to check if the device is staged.

        Returns:
            list:   List of objects that were staged, i.e. [self]
                    For devices with inheritance from ophyd, return
                    return super().stage() in the child class.
        """

    def unstage(self) -> list[object]:
        """Unstage method to cleanup after the acquisition.

        It can also be used to implement checks whether the acquisition was successful,
        inform BEC that the file has been succesfully written, or raise upon receiving
        feedback that the scan did not finish successful.

        Unstaging is not idempotent. If called twice it should simply resolve.
        It is recommended to return super().unstage() in the child class, if
        the child class also inherits from ophyd repository.
        """

    def stop(self, success: bool) -> None:
        """Stop method to stop the device.

        Args:
            success: Flag to indicate if the scan was successful or not.

        This method should be called to stop the device. It is recommended to call
        super().stop(success=success) if class inherits from ophyd repository.
        """


@runtime_checkable
class BECMixinProtocol(Protocol):
    """Protocol that offers BEC specific utility functionality for detectors."""

    USER_ACCESS: list[str]
    """
    List of methods/properties that will be exposed to the client interface in addition
    to the the already exposed signals, methods and properties.
    """

    scaninfo: bec_scaninfo_mixin
    """ 
    BEC scan info mixin class that provides an transparent Protocol to scan parameter 
    as provided by BEC. It is recommended to use this Protocol to retrieve scaninfo from Redis. 
    """

    stopped: bool
    """ 
    Flag to indicate if the device is stopped. 
    
    The stop method should set this flag to True, and i.e. stage to set it to False. 
    """

    filewriter: FileWriter
    """
    The file writer mixin main purpose is to unify and centralize the creation of 
    file paths within BEC. Therefore, we recommend devices to use the same mixin for creation of paths.
    """

    def pre_scan(self):
        """Pre-scan method is called from BEC right before executing scancore, thus
        right before the start of an acquisition.

        It can be used to trigger time critical functions from the device, which
        are prone to run into timeouts in case called too early.
        """


@runtime_checkable
class BECPositionerProtocol(Protocol):
    """Protocol with functionality specific for positioners in BEC."""

    @property
    def limits(self) -> tuple[float, float]:
        """Limits property for positioners.
        For an EpicsMotor, BEC will automatically recover the limits from the IOC.

        If not set, it returns (0,0).
        Note, low_limit = high_limit is equivalent to NO limits!

        Returns:
            tuple: Tuple with lower and upper limits
        """

    @property
    def low_limit(self) -> float:
        """Low limit property for positioners.

        Returns:
            float: Lower limit
        """

    @property
    def high_limit(self) -> float:
        """High limit property for positioners.

        Returns:
            float: Upper limit
        """

    def check_value(self, value: float) -> None:
        """Check whether value is within limits

        Args:
            value: value to check

        Raises:
            LimitError in case the requested motion is not inside of limits.
        """

    def move(self, position: float) -> DeviceStatus:
        """Move method for positioners.
        The returned DeviceStatus is marked as done once the positioner has reached the target position.
        DeviceStatus.wait() can be used to block until the move is completed.

        Args:
            position: position to move to

        Returns:
            DeviceStatus: DeviceStatus object
        """

    def set(self, position: float) -> DeviceStatus:
        """Set method for positioners.

        In principle, a set command is the same as move. This comes from ophyd upstream.
        We will have to review whether BEC requires both.

        Args:
            position: position to move to

        Returns:
            DeviceStatus: DeviceStatus object
        """


@runtime_checkable
class BECFlyerProtocol(BECScanProtocol, Protocol):
    """Protocol with functionality specific for flyers in BEC."""

    # def configure(self, d: dict):
    #     """Configure method of the flyer.
    #     It is an optional method, but does not need to be implemented by a flyer.
    #     Instead, stage can be used to prepare time critical operations on the device in preparation of a scan.

    #     Method to configure the flyer in preparation of a scan.

    #     Args:
    #         d (dict): Dictionary with configuration parameters, i.e. key value pairs of signal_name : value
    #     """

    def kickoff(self) -> DeviceStatus:
        """Kickoff method for flyers.

        The returned DeviceStatus is marked as done once the flyer start flying,
        i.e. is ready to be triggered.

        Returns:
            DeviceStatus: DeviceStatus object
        """

    def complete(self) -> DeviceStatus:
        """Complete method for flyers.

        The returned DeviceStatus is marked as done once the flyer has completed.

        Returns:
            DeviceStatus: DeviceStatus object
        """


@runtime_checkable
class BECRotationProtocol(Protocol):
    """Protocol which defines functionality for a tomography stage for ophyd devices"""

    allow_mod360: Component
    """Signal to define whether mod360 operations are allowed. """

    @property
    def has_mod360(self) -> bool:
        """Property to check if the motor has mod360 option

        Returns:
            bool: True if mod360 is possible on device, False otherwise
        """

    @property
    def has_freerun(self) -> bool:
        """Property to check if the motor has freerun option

        Returns:
            bool: True if freerun is allowed, False otherwise
        """

    @property
    def valid_rotation_modes(self) -> list[str]:
        """Method to get the valid rotation modes for the implemented motor.

        Returns:
            list: List of strings with valid rotation modes
        """

    def apply_mod360(self) -> None:
        """Method to apply the modulus 360 operation on the specific device.

        Childrens should override this method
        """


@runtime_checkable
class BECEventProtocol(Protocol):
    """Protocol for events in BEC.

    This is a first draft for the event protocol introduced throughout BEC.
    It needs to be review and extended before it can be used in production.
    """

    _callbacks: dict[dict]

    @property
    def event_types(self) -> tuple[str]:
        """Event types property"""

    def _run_subs(self, sub_type: str, **kwargs):
        """Run subscriptions for the event.

        Args:
            sub_type: Subscription type
            kwargs: Keyword arguments
        """

    def subscribe(self, callback: callable, event_type: str = None, run: bool = True):
        """Subscribe to the event.

        Args:
            callback (callable) :   Callback function
                                    The expected callback structure is:
                                    def cb(*args, obj:OphydObject, sub_type:str, **kwargs) -> None:
                                        pass
            event_type (str)    :   Event type, if None it defaults to obj._default_sub
                                    This maps to sub_type in _run_subs
            run (bool)          :   If true, run the callback directly.

        Returns:
            cid (int):              Callback id
        """

    def clear_sub(self, cb: callable, event_type: str = None):
        """Clear subscription, given the origianl callback fucntion

        Args:
            cb (callable)   : Callback
            event_type (str): Event type, if None it will be remove from all event_types
        """

    def unsubscribe(self, cid: int):
        """Unsubscribe from the event.

        Args:
            cid (int): Callback id
        """



# Standard ophyd classes
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO
from ophyd.quadem import QuadEM
from ophyd.sim import SynAxis, SynPeriodicSignal, SynSignal

from .aerotech.AerotechAutomation1 import (
    EpicsMotorX,
    aa1AxisDriveDataCollection,
    aa1AxisPsoDistance,
    aa1Controller,
    aa1GlobalVariableBindings,
    aa1GlobalVariables,
    aa1Tasks,
)
from .delay_generator_csaxs import DelayGeneratorcSAXS
from .eiger9m_csaxs import Eiger9McSAXS

# cSAXS
from .epics_motor_ex import EpicsMotorEx
from .falcon_csaxs import FalconcSAXS
from .flomni_sample_storage import FlomniSampleStorage
from .grashopper_tomcat import GrashopperTOMCAT
from .InsertionDevice import InsertionDevice
from .mcs_csaxs import MCScSAXS
from .pilatus_csaxs import PilatuscSAXS
from .psi_detector_base import CustomDetectorMixin, PSIDetectorBase
from .slits import SlitH, SlitV
from .specMotors import (
    Bpm4i,
    EnergyKev,
    GirderMotorPITCH,
    GirderMotorROLL,
    GirderMotorX1,
    GirderMotorY1,
    GirderMotorYAW,
    MonoTheta1,
    MonoTheta2,
    PmDetectorRotation,
    PmMonoBender,
)
from .SpmBase import SpmBase



import enum
import json
import os
import time

import numpy as np
import requests
from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import ADComponent as ADCpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV, Staged

from ophyd_devices.epics.devices.psi_detector_base import CustomDetectorMixin, PSIDetectorBase

logger = bec_logger.logger

MIN_READOUT = 3e-3


class PilatusError(Exception):
    """Base class for exceptions in this module."""


class PilatusTimeoutError(PilatusError):
    """Raised when the Pilatus does not respond in time during unstage."""


class TriggerSource(enum.IntEnum):
    """Trigger source options for the detector"""

    INTERNAL = 0
    EXT_ENABLE = 1
    EXT_TRIGGER = 2
    MULTI_TRIGGER = 3
    ALGINMENT = 4


class SLSDetectorCam(Device):
    """SLS Detector Camera - Pilatus

    Base class to map EPICS PVs to ophyd signals.
    """

    num_images = ADCpt(EpicsSignalWithRBV, "NumImages")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumExposures")
    delay_time = ADCpt(EpicsSignalWithRBV, "NumExposures")
    trigger_mode = ADCpt(EpicsSignalWithRBV, "TriggerMode")
    acquire = ADCpt(EpicsSignal, "Acquire")
    armed = ADCpt(EpicsSignalRO, "Armed")

    read_file_timeout = ADCpt(EpicsSignal, "ImageFileTmot")
    detector_state = ADCpt(EpicsSignalRO, "StatusMessage_RBV")
    status_message_camserver = ADCpt(EpicsSignalRO, "StringFromServer_RBV", string=True)
    acquire_time = ADCpt(EpicsSignal, "AcquireTime")
    acquire_period = ADCpt(EpicsSignal, "AcquirePeriod")
    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    file_path = ADCpt(EpicsSignalWithRBV, "FilePath")
    file_name = ADCpt(EpicsSignalWithRBV, "FileName")
    file_number = ADCpt(EpicsSignalWithRBV, "FileNumber")
    auto_increment = ADCpt(EpicsSignalWithRBV, "AutoIncrement")
    file_template = ADCpt(EpicsSignalWithRBV, "FileTemplate")
    file_format = ADCpt(EpicsSignalWithRBV, "FileNumber")
    gap_fill = ADCpt(EpicsSignalWithRBV, "GapFill")


class PilatusSetup(CustomDetectorMixin):
    """Pilatus setup class for cSAXS

    Parent class: CustomDetectorMixin

    """

    def initialize_default_parameter(self) -> None:
        """Set default parameters for Eiger9M detector"""
        self.update_readout_time()

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)

    def initialize_detector(self) -> None:
        """Initialize detector"""
        # Stops the detector
        self.stop_detector()
        # Sets the trigger source to GATING
        self.parent.set_trigger(TriggerSource.EXT_ENABLE)

    def prepare_detector(self) -> None:
        """
        Prepare detector for scan.

        Includes checking the detector threshold,
        setting the acquisition parameters and setting the trigger source
        """
        self.set_detector_threshold()
        self.set_acquisition_params()
        self.parent.set_trigger(TriggerSource.EXT_ENABLE)

    def set_detector_threshold(self) -> None:
        """
        Set correct detector threshold to 1/2 of current X-ray energy, allow 5% tolerance

        Threshold might be in ev or keV
        """

        # get current beam energy from device manageer
        mokev = self.parent.device_manager.devices.mokev.obj.read()[
            self.parent.device_manager.devices.mokev.name
        ]["value"]
        factor = 1

        # Check if energies are eV or keV, assume keV as the default
        unit = getattr(self.parent.cam.threshold_energy, "units", None)
        if unit is not None and unit == "eV":
            factor = 1000

        # set energy on detector
        setpoint = int(mokev * factor)

        # set threshold on detector
        threshold = self.parent.cam.threshold_energy.read()[self.parent.cam.threshold_energy.name][
            "value"
        ]
        if not np.isclose(setpoint / 2, threshold, rtol=0.05):
            self.parent.cam.threshold_energy.set(setpoint / 2)

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for the detector"""

        # Set number of images and frames (frames is for internal burst of detector)
        self.parent.cam.num_images.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.cam.num_frames.put(1)

        # Update the readout time of the detector
        self.update_readout_time()

    def create_directory(self, filepath: str) -> None:
        """Create directory if it does not exist"""
        os.makedirs(filepath, exist_ok=True)

    def stop_detector_backend(self) -> None:
        """Stop the file writer zmq service for pilatus_2"""
        self.close_file_writer()
        time.sleep(0.1)
        self.stop_file_writer()
        time.sleep(0.1)

    def close_file_writer(self) -> None:
        """
        Close the file writer for pilatus_2

        Delete the data from x12sa-pd-2

        """
        url = "http://x12sa-pd-2:8080/stream/pilatus_2"
        try:
            res = self.send_requests_delete(url=url)
            if not res.ok:
                res.raise_for_status()
        except Exception as exc:
            logger.info(f"Pilatus2 close threw Exception: {exc}")

    def stop_file_writer(self) -> None:
        """
        Stop the file writer for pilatus_2

        Runs on xbl-daq-34
        """
        url = "http://xbl-daq-34:8091/pilatus_2/stop"
        res = self.send_requests_put(url=url)
        if not res.ok:
            res.raise_for_status()

    def prepare_data_backend(self) -> None:
        """
        Prepare the detector backend of pilatus for a scan

        A zmq service is running on xbl-daq-34 that is waiting
        for a zmq message to start the writer for the pilatus_2 x12sa-pd-2

        """

        self.stop_detector_backend()

        self.parent.filepath = self.parent.filewriter.compile_full_filename("pilatus_2.h5")
        self.parent.cam.file_path.put("/dev/shm/zmq/")
        self.parent.cam.file_name.put(
            f"{self.parent.scaninfo.username}_2_{self.parent.scaninfo.scan_number:05d}"
        )
        self.parent.cam.auto_increment.put(1)  # auto increment
        self.parent.cam.file_number.put(0)  # first iter
        self.parent.cam.file_format.put(0)  # 0: TIFF
        self.parent.cam.file_template.put("%s%s_%5.5d.cbf")

        # TODO better to remove hard coded path with link to home directory/pilatus_2
        basepath = f"/sls/X12SA/data/{self.parent.scaninfo.username}/Data10/pilatus_2/"
        self.parent.filepath_raw = os.path.join(
            basepath,
            self.parent.filewriter.get_scan_directory(self.parent.scaninfo.scan_number, 1000, 5),
        )
        # Make directory if needed
        self.create_directory(self.parent.filepath_raw)

        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        # start the stream on x12sa-pd-2
        url = "http://x12sa-pd-2:8080/stream/pilatus_2"
        data_msg = {
            "source": [
                {
                    "searchPath": "/",
                    "searchPattern": "glob:*.cbf",
                    "destinationPath": self.parent.filepath_raw,
                }
            ]
        }
        res = self.send_requests_put(url=url, data=data_msg, headers=headers)
        logger.info(f"{res.status_code} -  {res.text} - {res.content}")

        if not res.ok:
            res.raise_for_status()

        # start the data receiver on xbl-daq-34
        url = "http://xbl-daq-34:8091/pilatus_2/run"
        data_msg = [
            "zmqWriter",
            self.parent.scaninfo.username,
            {
                "addr": "tcp://x12sa-pd-2:8888",
                "dst": ["file"],
                "numFrm": int(
                    self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                ),
                "timeout": 2000,
                "ifType": "PULL",
                "user": self.parent.scaninfo.username,
            },
        ]
        res = self.send_requests_put(url=url, data=data_msg, headers=headers)
        logger.info(f"{res.status_code}  - {res.text} - {res.content}")

        if not res.ok:
            res.raise_for_status()

        # Wait for server to become available again
        time.sleep(0.1)
        logger.info(f"{res.status_code} -{res.text} - {res.content}")

        # Send requests.put to xbl-daq-34 to wait for data
        url = "http://xbl-daq-34:8091/pilatus_2/wait"
        data_msg = [
            "zmqWriter",
            self.parent.scaninfo.username,
            {
                "frmCnt": int(
                    self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                ),
                "timeout": 2000,
            },
        ]
        try:
            res = self.send_requests_put(url=url, data=data_msg, headers=headers)
            logger.info(f"{res}")

            if not res.ok:
                res.raise_for_status()
        except Exception as exc:
            logger.info(f"Pilatus2 wait threw Exception: {exc}")

    def send_requests_put(self, url: str, data: list = None, headers: dict = None) -> object:
        """
        Send a put request to the given url

        Args:
            url (str): url to send the request to
            data (dict): data to be sent with the request (optional)
            headers (dict): headers to be sent with the request (optional)

        Returns:
            status code of the request
        """
        return requests.put(url=url, data=json.dumps(data), headers=headers, timeout=5)

    def send_requests_delete(self, url: str, headers: dict = None) -> object:
        """
        Send a delete request to the given url

        Args:
            url (str): url to send the request to
            headers (dict): headers to be sent with the request (optional)

        Returns:
            status code of the request
        """
        return requests.delete(url=url, headers=headers, timeout=5)

    def pre_scan(self) -> None:
        """
        Pre_scan function call

        This function is called just before the scan core.
        Here it is used to arm the detector for the acquisition

        """
        self.arm_acquisition()

    def arm_acquisition(self) -> None:
        """Arms the detector for the acquisition"""
        self.parent.cam.acquire.put(1)
        # TODO is this sleep needed? to be tested with detector and for how long
        time.sleep(0.5)

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the filepath to REDIS and publish the event for the h5_converter

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
        """
        pipe = self.parent.connector.pipeline()
        if successful is None:
            msg = messages.FileMessage(
                file_path=self.parent.filepath,
                done=done,
                metadata={"input_path": self.parent.filepath_raw},
            )
        else:
            msg = messages.FileMessage(
                file_path=self.parent.filepath,
                done=done,
                successful=successful,
                metadata={"input_path": self.parent.filepath_raw},
            )
        self.parent.connector.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scan_id, self.parent.name),
            msg,
            pipe=pipe,
        )
        self.parent.connector.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg, pipe=pipe
        )
        pipe.execute()

    def finished(self) -> None:
        """Check if acquisition is finished."""
        # pylint: disable=protected-access
        # TODO: at the moment this relies on device.mcs.obj._staged attribute
        signal_conditions = [
            (lambda: self.parent.device_manager.devices.mcs.obj._staged, Staged.no)
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            raise PilatusTimeoutError(
                f"Reached timeout with detector state {signal_conditions[0][0]}, std_daq state"
                f" {signal_conditions[1][0]} and received frames of {signal_conditions[2][0]} for"
                " the file writer"
            )
        self.stop_detector()
        self.stop_detector_backend()

    def stop_detector(self) -> None:
        """Stop detector"""
        self.parent.cam.acquire.put(0)

    def check_scan_id(self) -> None:
        """Checks if scan_id has changed and stops the scan if it has"""
        old_scan_id = self.parent.scaninfo.scan_id
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scan_id != old_scan_id:
            self.parent.stopped = True


class PilatuscSAXS(PSIDetectorBase):
    """Pilatus_2 300k detector for CSAXS

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (Eiger9MSetup)   : Custom detector setup class for cSAXS,
                                              inherits from CustomDetectorMixin
        cam (SLSDetectorCam)                : Detector camera
        MIN_READOUT (float)                 : Minimum readout time for the detector

    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe"]

    # specify Setup class
    custom_prepare_cls = PilatusSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3
    # specify class attributes
    cam = ADCpt(SLSDetectorCam, "cam1:")

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source for the detector"""
        value = trigger_source
        self.cam.trigger_mode.put(value)


if __name__ == "__main__":
    pilatus_2 = PilatuscSAXS(name="pilatus_2", prefix="X12SA-ES-PILATUS300K:", sim_mode=True)



import inspect

import lmfit


class LmfitModelMixin:

    # def __init__(self):
    #     self.model = lmfit.models.GaussianModel()
    #     self.params = self.model.make_params()
    #     self.params["center"].set(value=0)
    #     self.params["amplitude"].set(value=1)
    #     self.params["sigma"].set(value=1)

    @staticmethod
    def available_models() -> dict:
        """
        Get available models from lmfit.models.

        Exclude Gaussian2dModel, ExpressionModel, Model, SplineModel.
        """
        avail_models = {}
        for name, model_cls in inspect.getmembers(lmfit.models):
            try:
                is_model = issubclass(model_cls, lmfit.model.Model)
            except TypeError:
                is_model = False
            if is_model and name not in [
                "Gaussian2dModel",
                "ExpressionModel",
                "Model",
                "SplineModel",
            ]:
                avail_models[name] = model_cls
        return avail_models

    def create_properties(self):
        """
        Create properties for model parameters.
        """
        for name in self.available_models():
            setattr(self, name, param)

    @staticmethod
    def get_model(model: str) -> lmfit.Model:
        """Get model for given string."""
        if isinstance(model, str):
            model = getattr(lmfit.models, model, None)
        if not model:
            raise ValueError(f"Model {model} not found.")
        return model



from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO


class SequencerX12SA(Device):
    """Sequencer for flyscans with epics motor controller and owis stages"""

    desired_output_link_1 = Cpt(Signal, "DOL1")
    desired_output_value_1 = Cpt(EpicsSignal, "DO1")
    output_link_1 = Cpt(EpicsSignal, "LNK1")
    delay_time_1 = Cpt(EpicsSignal, "DLY1")
    desired_output_link_2 = Cpt(EpicsSignal, "DOL2")
    desired_output_value_2 = Cpt(EpicsSignal, "DO2")
    output_link_2 = Cpt(EpicsSignal, "LNK2")
    delay_time_2 = Cpt(EpicsSignal, "DLY2")

    select_mechanism = Cpt(EpicsSignal, "SELM", string=True)
    link_selection = Cpt(EpicsSignal, "SELN")
    process = Cpt(EpicsSignal, "PROC", string=True)

    status = Cpt(EpicsSignalRO, "STAT", string=True)
    processing_active = Cpt(EpicsSignalRO, "PACT")

    # def __init__(
    #     self,
    #     prefix="",
    #     *,
    #     name,
    #     kind=None,
    #     read_attrs=None,
    #     configuration_attrs=None,
    #     parent=None,
    #     **kwargs
    # ):
    # # get configuration attributes from kwargs and then remove them
    # attrs = {}
    # for key, value in kwargs.items():
    #     if hasattr(EpicsMotorEx, key) and isinstance(getattr(EpicsMotorEx, key), Cpt):
    #         attrs[key] = value
    # for key in attrs:
    #     kwargs.pop(key)

    # super().__init__(
    #     prefix,
    #     name=name,
    #     kind=kind,
    #     read_attrs=read_attrs,
    #     configuration_attrs=configuration_attrs,
    #     parent=parent,
    #     **kwargs
    # )

    # # set configuration attributes
    # for key, value in attrs.items():
    #     # print out attributes that are being configured
    #     print("setting ", key, "=", value)
    #     getattr(self, key).put(value)
